from activitysim.benchmarking.componentwise import (
    template_component_timings,
    template_setup_cache,
)

EXAMPLE_NAME = "prototype_mtc_full"
CONFIGS_DIRS = ("configs",)
DYNAMIC_CONFIG_DIR = "bench_configs"
DATA_DIR = "data"
OUTPUT_DIR = "output"
COMPONENT_NAMES = [
    # "compute_accessibility",
    "school_location",
    "workplace_location",
    "auto_ownership_simulate",
    "free_parking",
    "cdap_simulate",
    "mandatory_tour_frequency",
    "mandatory_tour_scheduling",
    "joint_tour_frequency",
    "joint_tour_composition",
    "joint_tour_participation",
    "joint_tour_destination",
    "joint_tour_scheduling",
    "non_mandatory_tour_frequency",
    "non_mandatory_tour_destination",
    "non_mandatory_tour_scheduling",
    "tour_mode_choice_simulate",
    "atwork_subtour_frequency",
    "atwork_subtour_destination",
    "atwork_subtour_scheduling",
    "atwork_subtour_mode_choice",
    "stop_frequency",
    "trip_purpose",
    "trip_destination",
    "trip_purpose_and_destination",
    "trip_scheduling",
    "trip_mode_choice",
    # "write_data_dictionary",
    # "track_skim_usage",
    "write_trip_matrices",
    # "write_tables",
]
BENCHMARK_SETTINGS = {
    "households_sample_size": 48_769,
}
SKIM_CACHE = False
PRELOAD_INJECTABLES = ("skim_dict",)
REPEAT = 1
NUMBER = 1
TIMEOUT = 36000.0  # ten hours
VERSION = "1"


def setup_cache():
    template_setup_cache(
        EXAMPLE_NAME,
        COMPONENT_NAMES,
        BENCHMARK_SETTINGS,
        dict(
            read_skim_cache=SKIM_CACHE,
            write_skim_cache=SKIM_CACHE,
        ),
        CONFIGS_DIRS,
        DATA_DIR,
        OUTPUT_DIR,
        config_overload_dir=DYNAMIC_CONFIG_DIR,
    )


template_component_timings(
    globals(),
    COMPONENT_NAMES,
    EXAMPLE_NAME,
    (DYNAMIC_CONFIG_DIR, *CONFIGS_DIRS),
    DATA_DIR,
    OUTPUT_DIR,
    PRELOAD_INJECTABLES,
    REPEAT,
    NUMBER,
    TIMEOUT,
    VERSION,
)



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd

from activitysim.core import chunk, config, inject, logit, pipeline, simulate, tracing

logger = logging.getLogger(__name__)


def _clip_probs(choosers_df, probs, depart_alt_base):
    """
    zero out probs before chooser.earliest or after chooser.latest

    Parameters
    ----------
    choosers_df: pd.DataFrame
    probs: pd.DataFrame
        one row per chooser, one column per time period, with float prob of picking that time period

    depart_alt_base: int
        int to add to probs column index to get time period it represents.
        e.g. depart_alt_base = 5 means first column (column 0) represents 5 am

    Returns
    -------
    probs: pd.DataFrame
        clipped version of probs

    """

    # there should be one row in probs per trip
    assert choosers_df.shape[0] == probs.shape[0]

    # probs should sum to 1 across rows before clipping
    probs = probs.div(probs.sum(axis=1), axis=0)

    num_rows, num_cols = probs.shape
    ix_map = (
        np.tile(np.arange(0, num_cols), num_rows).reshape(num_rows, num_cols)
        + depart_alt_base
    )
    # 5 6 7 8 9 10...
    # 5 6 7 8 9 10...
    # 5 6 7 8 9 10...

    clip_mask = (
        (ix_map >= choosers_df.earliest.values.reshape(num_rows, 1))
        & (ix_map <= choosers_df.latest.values.reshape(num_rows, 1))
    ) * 1
    #  [0 0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 0 0 0]
    #  [0 0 0 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 0]
    #  [0 0 0 0 0 0 0 0 0 1 0 0 0 0 0 0 0 0 0]...

    probs = probs * clip_mask

    return probs


def _report_bad_choices(bad_row_map, df, filename, trace_label, trace_choosers=None):
    """

    Parameters
    ----------
    bad_row_map
    df : pandas.DataFrame
        utils or probs dataframe
    trace_choosers : pandas.dataframe
        the choosers df (for interaction_simulate) to facilitate the reporting of hh_id
        because we  can't deduce hh_id from the interaction_dataset which is indexed on index
        values from alternatives df

    """

    df = df[bad_row_map]
    if trace_choosers is None:
        hh_ids = tracing.hh_id_for_chooser(df.index, df)
    else:
        hh_ids = tracing.hh_id_for_chooser(df.index, trace_choosers)
    df["household_id"] = hh_ids

    filename = "%s.%s" % (trace_label, filename)

    logger.info("dumping %s" % filename)
    tracing.write_csv(df, file_name=filename, transpose=False)

    # log the indexes of the first MAX_PRINT offending rows
    MAX_PRINT = 0
    for idx in df.index[:MAX_PRINT].values:

        row_msg = "%s : failed %s = %s (hh_id = %s)" % (
            trace_label,
            df.index.name,
            idx,
            df.household_id.loc[idx],
        )

        logger.warning(row_msg)


def _preprocess_departure_probs(
    choosers_df,
    choosers,
    probs_spec,
    probs_join_cols,
    clip_earliest_latest,
    depart_alt_base,
    first_trip_in_leg,
):

    # zero out probs outside earliest-latest window if one exists
    probs_cols = [c for c in probs_spec.columns if c not in probs_join_cols]
    if clip_earliest_latest:
        chooser_probs = _clip_probs(choosers_df, choosers[probs_cols], depart_alt_base)
    else:
        chooser_probs = choosers.loc[:, probs_cols]

    if first_trip_in_leg:
        # probs should sum to 1 unless all zero
        chooser_probs = chooser_probs.div(chooser_probs.sum(axis=1), axis=0).fillna(0)

    return chooser_probs


def _preprocess_stop_duration_probs(choosers):

    # convert wide to long. duration probs are stored in long format so that
    # inbound/outbound duration probs, which both have a 0 alternative, can be
    # stored in a single config file. open to better suggestions here.
    choosers["periods_left"] = choosers["latest"] - choosers["earliest"]
    choosers = choosers.query("periods_left_min <= periods_left <= periods_left_max")
    chooser_probs = choosers.reset_index().pivot(
        index="trip_id", columns="duration_offset", values="prob"
    )

    # re-order columns if using negative offsets bc pivot (above) will sort
    # columns in ascending order by default. but we need the zero-indexed
    # col to be the 0 offset alternative for both outbound (positive) and
    # inbound (negative) alts in order for the post-processing step to work.
    if all(chooser_probs.columns < 1):
        chooser_probs = chooser_probs.sort_index(axis=1, ascending=False)

    return chooser_probs


def _preprocess_scheduling_probs(
    scheduling_mode,
    choosers_df,
    choosers,
    probs_spec,
    probs_join_cols,
    clip_earliest_latest,
    depart_alt_base,
    first_trip_in_leg,
):
    """
    Preprocesses the choosers tables depending on the trip scheduling mode
    selected.
    """

    if scheduling_mode == "departure":
        chooser_probs = _preprocess_departure_probs(
            choosers_df,
            choosers,
            probs_spec,
            probs_join_cols,
            clip_earliest_latest,
            depart_alt_base,
            first_trip_in_leg,
        )
    elif scheduling_mode == "stop_duration":
        chooser_probs = _preprocess_stop_duration_probs(choosers)
    else:
        logger.error(
            "Invalid scheduling mode specified: {0}.".format(scheduling_mode),
            "Please select one of ['departure', 'stop_duration'] and try again.",
        )

    # probs should sum to 1 with residual probs resulting in choice of 'fail'
    chooser_probs["fail"] = 1 - chooser_probs.sum(axis=1).clip(0, 1)

    return chooser_probs


def _postprocess_scheduling_choices(
    scheduling_mode, depart_alt_base, choices, choice_cols, choosers
):

    """
    This method just converts the choice column indexes returned by the
    logit.make_choices() method into actual departure time values that are
    more useful for downstream models.
    """
    # convert alt choice index to depart time (setting failed choices to -1)
    failed = choices == choice_cols.get_loc("fail")

    # For the stop duration-based probabilities, the alternatives are offsets that
    # get applied to trip-specific departure and arrival times, so depart_alt_base
    # is a column/series rather than a scalar.
    if scheduling_mode == "stop_duration":

        # for outbound trips, offsets get added to the departure time constraint
        if choosers.outbound.all():
            depart_alt_base = choosers["earliest"]

        # for inbound trips, offsets get subtracted from tour end constraint
        elif not choosers.outbound.any():
            depart_alt_base = choosers["latest"]
            choices *= -1

        else:
            logger.error(
                "Outbound trips are being scheduled at the same "
                "time as inbound trips. That should never happen."
            )

    choices = (choices + depart_alt_base).where(~failed, -1)

    return choices, failed


def make_scheduling_choices(
    choosers_df,
    scheduling_mode,
    probs_spec,
    probs_join_cols,
    depart_alt_base,
    first_trip_in_leg,
    report_failed_trips,
    trace_hh_id,
    trace_label,
    trace_choice_col_name="depart",
    clip_earliest_latest=True,
):
    """
    We join each trip with the appropriate row in probs_spec by joining on probs_join_cols,
    which should exist in both trips, probs_spec dataframe.

    Parameters
    ----------
    choosers: pd.DataFrame
    scheduling_mode: str
        Either 'departure' or 'stop_duration' depending on whether the probability
        lookup table is keyed on depature period or stop duration.
    trips: pd.DataFrame
    probs_spec: pd.DataFrame
        Dataframe of probs for choice of depart times and join columns to match them with trips.
        Depart columns names are irrelevant. Instead, they are position dependent,
        time period choice is their index + depart_alt_base
    depart_alt_base: int
        int to add to probs column index to get time period it represents.
        e.g. depart_alt_base = 5 means first column (column 0) represents 5 am
    report_failed_trips : bool
    trace_hh_id
    trace_label

    Returns
    -------
    choices: pd.Series
        time periods depart choices, one per trip (except for trips with zero probs)
    """

    choosers = pd.merge(
        choosers_df.reset_index(), probs_spec, on=probs_join_cols, how="left"
    ).set_index(choosers_df.index.name)
    chunk.log_df(trace_label, "choosers", choosers)

    if trace_hh_id and tracing.has_trace_targets(choosers_df):
        tracing.trace_df(choosers, "%s.choosers" % trace_label)

    # different pre-processing is required based on the scheduling mode
    chooser_probs = _preprocess_scheduling_probs(
        scheduling_mode,
        choosers_df,
        choosers,
        probs_spec,
        probs_join_cols,
        clip_earliest_latest,
        depart_alt_base,
        first_trip_in_leg,
    )

    chunk.log_df(trace_label, "chooser_probs", chooser_probs)

    if trace_hh_id and tracing.has_trace_targets(choosers_df):
        tracing.trace_df(chooser_probs, "%s.chooser_probs" % trace_label)

    raw_choices, rands = logit.make_choices(
        chooser_probs, trace_label=trace_label, trace_choosers=choosers
    )

    chunk.log_df(trace_label, "choices", raw_choices)
    chunk.log_df(trace_label, "rands", rands)

    if trace_hh_id and tracing.has_trace_targets(choosers_df):
        tracing.trace_df(
            raw_choices,
            "%s.choices" % trace_label,
            columns=[None, trace_choice_col_name],
        )
        tracing.trace_df(rands, "%s.rands" % trace_label, columns=[None, "rand"])

    # different post-processing is required based on the scheduling mode
    choices, failed = _postprocess_scheduling_choices(
        scheduling_mode,
        depart_alt_base,
        raw_choices,
        chooser_probs.columns,
        choosers_df,
    )

    chunk.log_df(trace_label, "failed", failed)

    # report failed trips while we have the best diagnostic info
    if report_failed_trips and failed.any():
        _report_bad_choices(
            bad_row_map=failed,
            df=choosers,
            filename="failed_choosers",
            trace_label=trace_label,
            trace_choosers=None,
        )

    # trace before removing failures
    if trace_hh_id and tracing.has_trace_targets(choosers_df):
        tracing.trace_df(
            choices, "%s.choices" % trace_label, columns=[None, trace_choice_col_name]
        )
        tracing.trace_df(rands, "%s.rands" % trace_label, columns=[None, "rand"])

    # remove any failed choices
    if failed.any():
        choices = choices[~failed]

    if all([check_col in choosers_df.columns for check_col in ["earliest", "latest"]]):
        assert (choices >= choosers_df.earliest[~failed]).all()
        assert (choices <= choosers_df.latest[~failed]).all()

    return choices



import os

_directory = None


def get_dir():
    global _directory
    if _directory is None:
        _directory = os.environ.get("ASIM_ASV_WORKSPACE", None)
    if _directory is None:
        _directory = os.environ.get("ASV_CONF_DIR", None)
    return _directory


def set_dir(directory):
    global _directory
    if directory:
        _directory = directory
    else:
        _directory = os.environ.get("ASIM_ASV_WORKSPACE", None)



# ActivitySim
# See full license in LICENSE.txt.
import os
import subprocess

import pandas as pd
import pandas.testing as pdt
import pkg_resources
import pytest

from activitysim.core import inject


def teardown_function(func):
    inject.clear_cache()
    inject.reinject_decorated_tables()


def example_path(dirname):
    resource = os.path.join("examples", "placeholder_multiple_zone", dirname)
    return pkg_resources.resource_filename("activitysim", resource)


def mtc_example_path(dirname):
    resource = os.path.join("examples", "prototype_mtc", dirname)
    return pkg_resources.resource_filename("activitysim", resource)


def build_data():
    # FIXME this irks travis
    # subprocess.check_call(['coverage', 'run', example_path('scripts/two_zone_example_data.py')])
    # subprocess.check_call(['coverage', 'run', example_path('scripts/three_zone_example_data.py')])
    pass


@pytest.fixture(scope="module")
def data():
    build_data()


def run_test(zone, multiprocess=False):
    def test_path(dirname):
        return os.path.join(os.path.dirname(__file__), dirname)

    def regress(zone):

        # regress tours
        regress_tours_df = pd.read_csv(
            test_path(f"regress/final_tours_{zone}_zone.csv")
        )
        tours_df = pd.read_csv(test_path("output/final_tours.csv"))
        tours_df.to_csv(
            test_path(f"regress/final_tours_{zone}_zone_last_run.csv"), index=False
        )
        print(f"regress tours")
        pdt.assert_frame_equal(tours_df, regress_tours_df, rtol=1e-03)

        # regress trips
        regress_trips_df = pd.read_csv(
            test_path(f"regress/final_trips_{zone}_zone.csv")
        )
        trips_df = pd.read_csv(test_path("output/final_trips.csv"))
        trips_df.to_csv(
            test_path(f"regress/final_trips_{zone}_zone_last_run.csv"), index=False
        )
        print(f"regress trips")
        pdt.assert_frame_equal(trips_df, regress_trips_df, rtol=1e-03)

    file_path = os.path.join(os.path.dirname(__file__), "simulation.py")

    run_args = [
        "-c",
        test_path(f"configs_{zone}_zone"),
        "-c",
        example_path(f"configs_{zone}_zone"),
        "-c",
        mtc_example_path(f"configs"),
        "-d",
        example_path(f"data_{zone}"),
        "-o",
        test_path("output"),
    ]

    if multiprocess:
        run_args = run_args + ["-s", "settings_mp"]
    elif zone == "3":
        run_args = run_args + ["-s", "settings_static"]

    subprocess.run(["coverage", "run", "-a", file_path] + run_args, check=True)

    regress(zone)


def test_2_zone(data):
    run_test(zone="2", multiprocess=False)


def test_2_zone_mp(data):
    run_test(zone="2", multiprocess=True)


def test_3_zone(data):
    # python simulation.py -c configs_3_zone -c ../configs_3_zone -c \
    # ../../prototype_mtc/configs -d ../data_3 -o output -s settings_mp
    run_test(zone="3", multiprocess=False)


def test_3_zone_mp(data):
    run_test(zone="3", multiprocess=True)


if __name__ == "__main__":

    build_data()
    run_test(zone="2", multiprocess=False)
    run_test(zone="2", multiprocess=True)

    run_test(zone="3", multiprocess=False)
    run_test(zone="3", multiprocess=True)



# remove unused skims from full scale sandag cross border skim files to
# reduce file size for upload to GitHub since there's a 2GB file size limit
# run this script and then run these repack commands to reduce file size

# Ben.Stabler@rsginc.com, 10/18/21

# h5repack -i traffic_skims_xborder_EA.omx -o traffic_skims_xborder_EA_repacked.omx
# h5repack -i traffic_skims_xborder_AM.omx -o traffic_skims_xborder_AM_repacked.omx
# h5repack -i traffic_skims_xborder_MD.omx -o traffic_skims_xborder_MD_repacked.omx
# h5repack -i traffic_skims_xborder_PM.omx -o traffic_skims_xborder_PM_repacked.omx
# h5repack -i traffic_skims_xborder_EV.omx -o traffic_skims_xborder_EV_repacked.omx

import tables

time_periods = ["EA", "AM", "MD", "PM", "EV"]

skims_to_remove = [
    "/data/HOV2_H_HOVDIST",
    "/data/HOV2_H_REL",
    "/data/HOV2_H_TOLLDIST",
    "/data/HOV2_L_HOVDIST",
    "/data/HOV2_L_REL",
    "/data/HOV2_L_TOLLDIST",
    "/data/HOV2_M_HOVDIST",
    "/data/HOV2_M_REL",
    "/data/HOV2_M_TOLLDIST",
    "/data/HOV3_H_HOVDIST",
    "/data/HOV3_H_REL",
    "/data/HOV3_H_TOLLDIST",
    "/data/HOV3_L_HOVDIST",
    "/data/HOV3_L_REL",
    "/data/HOV3_L_TOLLDIST",
    "/data/HOV3_M_HOVDIST",
    "/data/HOV3_M_REL",
    "/data/HOV3_M_TOLLDIST",
    "/data/SOV_NT_H_REL",
    "/data/SOV_NT_H_TOLLDIST",
    "/data/SOV_NT_L_REL",
    "/data/SOV_NT_L_TOLLDIST",
    "/data/SOV_NT_M_REL",
    "/data/SOV_NT_M_TOLLDIST",
    "/data/SOV_TR_H_DIST",
    "/data/SOV_TR_H_REL",
    "/data/SOV_TR_H_TIME",
    "/data/SOV_TR_H_TOLLCOST",
    "/data/SOV_TR_H_TOLLDIST",
    "/data/SOV_TR_L_DIST",
    "/data/SOV_TR_L_REL",
    "/data/SOV_TR_L_TIME",
    "/data/SOV_TR_L_TOLLCOST",
    "/data/SOV_TR_L_TOLLDIST",
    "/data/SOV_TR_M_DIST",
    "/data/SOV_TR_M_REL",
    "/data/SOV_TR_M_TIME",
    "/data/SOV_TR_M_TOLLCOST",
    "/data/SOV_TR_M_TOLLDIST",
    "/data/TRK_H_DIST",
    "/data/TRK_H_TIME",
    "/data/TRK_H_TOLLCOST",
    "/data/TRK_L_DIST",
    "/data/TRK_L_TIME",
    "/data/TRK_L_TOLLCOST",
    "/data/TRK_M_DIST",
    "/data/TRK_M_TIME",
    "/data/TRK_M_TOLLCOST",
]

for time_period in time_periods:
    f = tables.open_file("traffic_skims_xborder_" + time_period + ".omx", "a")
    for skim in skims_to_remove:
        print(skim + "__" + time_period)
        f.remove_node(skim + "__" + time_period)
    f.close()



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd

from activitysim.core import config, inject, pipeline, simulate, tracing
from activitysim.core.util import assign_in_place

from .util import estimation, tour_destination

logger = logging.getLogger(__name__)


@inject.step()
def non_mandatory_tour_destination(
    tours, persons_merged, network_los, chunk_size, trace_hh_id
):

    """
    Given the tour generation from the above, each tour needs to have a
    destination, so in this case tours are the choosers (with the associated
    person that's making the tour)
    """

    trace_label = "non_mandatory_tour_destination"
    model_settings_file_name = "non_mandatory_tour_destination.yaml"
    model_settings = config.read_model_settings(model_settings_file_name)

    logsum_column_name = model_settings.get("DEST_CHOICE_LOGSUM_COLUMN_NAME")
    want_logsums = logsum_column_name is not None

    sample_table_name = model_settings.get("DEST_CHOICE_SAMPLE_TABLE_NAME")
    want_sample_table = (
        config.setting("want_dest_choice_sample_tables")
        and sample_table_name is not None
    )

    tours = tours.to_frame()

    persons_merged = persons_merged.to_frame()

    # choosers are tours - in a sense tours are choosing their destination
    non_mandatory_tours = tours[tours.tour_category == "non_mandatory"]

    # - if no mandatory_tours
    if non_mandatory_tours.shape[0] == 0:
        tracing.no_results(trace_label)
        return

    estimator = estimation.manager.begin_estimation("non_mandatory_tour_destination")
    if estimator:
        estimator.write_coefficients(model_settings=model_settings)
        # estimator.write_spec(model_settings, tag='SAMPLE_SPEC')
        estimator.write_spec(model_settings, tag="SPEC")
        estimator.set_alt_id(model_settings["ALT_DEST_COL_NAME"])
        estimator.write_table(
            inject.get_injectable("size_terms"), "size_terms", append=False
        )
        estimator.write_table(
            inject.get_table("land_use").to_frame(), "landuse", append=False
        )
        estimator.write_model_settings(model_settings, model_settings_file_name)

    choices_df, save_sample_df = tour_destination.run_tour_destination(
        non_mandatory_tours,
        persons_merged,
        want_logsums,
        want_sample_table,
        model_settings,
        network_los,
        estimator,
        chunk_size,
        trace_hh_id,
        trace_label,
    )

    if estimator:
        estimator.write_choices(choices_df.choice)
        choices_df.choice = estimator.get_survey_values(
            choices_df.choice, "tours", "destination"
        )
        estimator.write_override_choices(choices_df.choice)
        estimator.end_estimation()

    non_mandatory_tours["destination"] = choices_df.choice

    assign_in_place(tours, non_mandatory_tours[["destination"]])

    if want_logsums:
        non_mandatory_tours[logsum_column_name] = choices_df["logsum"]
        assign_in_place(tours, non_mandatory_tours[[logsum_column_name]])

    pipeline.replace_table("tours", tours)

    if want_sample_table:
        assert len(save_sample_df.index.get_level_values(0).unique()) == len(choices_df)
        # save_sample_df.set_index(model_settings['ALT_DEST_COL_NAME'], append=True, inplace=True)
        pipeline.extend_table(sample_table_name, save_sample_df)

    if trace_hh_id:
        tracing.trace_df(
            tours[tours.tour_category == "non_mandatory"],
            label="non_mandatory_tour_destination",
            slicer="person_id",
            index_label="tour",
            columns=None,
            warn_if_empty=True,
        )



# Creating the Two Zone Example Data
#
# Transform the TM1 TAZ-based model 25 zone inputs to a two-zone (MAZ and TAZ) set of inputs for software development.
#
# The 25 zones are downtown San Francisco and they are converted to 25 MAZs.
# MAZs 1,2,3,4 are small and adjacent and assigned TAZ 2 and TAP 10002.
# MAZs 13,14,15 are small and adjacent and as signed TAZ 14 and TAP 10014.
# TAZs 1,3,4,13,15 are removed from the final data set.
#
# This script should work for the full TM1 example as well.

import os
import sys

import numpy as np
import openmatrix as omx
import pandas as pd

# Create example directory

input_data = os.path.join(os.path.dirname(__file__), "../data_1")
output_data = os.path.join(os.path.dirname(__file__), "../data_2")
MAZ_MULTIPLIER = 1000

# ### initialize output data directory

# new empty output_dir
if os.path.exists(output_data):
    # shutil.rmtree(output_data)
    # os.makedirs(output_data)
    file_type = ("csv", "omx")
    for file_name in os.listdir(output_data):
        if file_name.endswith(file_type):
            os.unlink(os.path.join(output_data, file_name))
else:
    os.makedirs(output_data)

# ### Convert tazs to mazs and add transit access distance by mode

land_use = pd.read_csv(os.path.join(input_data, "land_use.csv"))

if "ZONE" in land_use.columns:
    land_use.insert(loc=0, column="MAZ", value=land_use.ZONE)
    land_use.insert(loc=1, column="TAZ", value=land_use.ZONE)
    land_use.drop(columns=["ZONE"], inplace=True)
else:
    land_use.insert(loc=0, column="MAZ", value=land_use.TAZ)

land_use.TAZ = land_use.TAZ.replace([1, 2, 3, 4], 2)
land_use.TAZ = land_use.TAZ.replace([13, 14, 15], 14)

# make MAZ indexes different from TAZ to drive MAZ/TAZ confusion errors and omisisons
land_use.MAZ *= MAZ_MULTIPLIER

shortWalk = 0.333  # the tm1 example assumes this distance for transit access
longWalk = 0.667
land_use["access_dist_transit"] = shortWalk

# FIXME - could assign longWalk where maz != taz, but then results wodl differe from one-zone
# land_use['access_dist_transit'] =\
#     np.where(land_use.TAZ*MAZ_MULTIPLIER==land_use.MAZ, shortWalk, longWalk)


land_use.to_csv(os.path.join(output_data, "land_use.csv"), index=False)

# ### Put households in mazs instead of tazs

households = pd.read_csv(os.path.join(input_data, "households.csv"))
households.rename(columns={"TAZ": "MAZ"}, inplace=True)
households.MAZ *= MAZ_MULTIPLIER
households.to_csv(os.path.join(output_data, "households.csv"), index=False)

persons = pd.read_csv(os.path.join(input_data, "persons.csv"))
persons.to_csv(os.path.join(output_data, "persons.csv"), index=False)

# ### Create maz correspondence file

# FIXME - not clear we need this
maz_df = land_use[["MAZ", "TAZ"]]
maz_df.to_csv(os.path.join(output_data, "maz.csv"), index=False)
print("maz.csv\n%s" % (maz_df.head(6),))

# ### Create taz file

# TAZ
# 2
# 5
# 6
# 7

new_zone_labels = np.unique(land_use.TAZ)
new_zone_indexes = new_zone_labels - 1
taz_df = pd.DataFrame({"TAZ": new_zone_labels}, index=new_zone_indexes)
taz_df.to_csv(os.path.join(output_data, "taz.csv"), index=False)
print("taz.csv\n%s" % (taz_df.head(6),))

# currently this has only the one TAZ column, but the legacy table had:
# index TAZ
# offset             int64
# terminal_time    float64  # occasional small integer (1-5), but mostly blank (only if it has a TAP?
# ptype            float64  # parking type at TAP? (rarer than terminal_time, never alone)

# ### Create taz skims

with omx.open_file(
    os.path.join(input_data, "skims.omx"), "r"
) as skims_file, omx.open_file(
    os.path.join(output_data, "taz_skims.omx"), "w"
) as output_skims_file:

    skims = skims_file.list_matrices()
    num_zones = skims_file.shape()[0]

    # assume zones labels were 1-based in skims file
    assert not skims_file.listMappings()
    assert num_zones == len(land_use)

    for skim_name in skims_file.list_matrices():

        old_skim = skims_file[skim_name][:]
        new_skim = old_skim[new_zone_indexes, :][:, new_zone_indexes]
        output_skims_file[skim_name] = new_skim
        # print("skim:", skim_name, ": shape", str(new_skim.shape))

    output_skims_file.create_mapping("taz", new_zone_labels)

print("taz skims created: " + os.path.join(output_data, "taz_skims.omx"))

# ### Create maz to maz time/distance

max_distance_for_walk = 1.0
max_distance_for_bike = 5.0


with omx.open_file(os.path.join(input_data, "skims.omx")) as skims_file:

    # create df with DIST column
    maz_to_maz = pd.DataFrame(np.transpose(skims_file["DIST"])).unstack().reset_index()
    maz_to_maz.columns = ["OMAZ", "DMAZ", "DIST"]
    maz_to_maz["OMAZ"] = (maz_to_maz["OMAZ"] + 1) * MAZ_MULTIPLIER
    maz_to_maz["DMAZ"] = (maz_to_maz["DMAZ"] + 1) * MAZ_MULTIPLIER

    # additional columns
    for c in ["DISTBIKE", "DISTWALK"]:
        maz_to_maz[c] = pd.DataFrame(np.transpose(skims_file[c])).unstack().values

    maz_to_maz.loc[
        maz_to_maz["DIST"] <= max_distance_for_walk, ["OMAZ", "DMAZ", "DISTWALK"]
    ].to_csv(os.path.join(output_data, "maz_to_maz_walk.csv"), index=False)

    maz_to_maz.loc[
        maz_to_maz["DIST"] <= max_distance_for_bike,
        ["OMAZ", "DMAZ", "DIST", "DISTBIKE"],
    ].to_csv(os.path.join(output_data, "maz_to_maz_bike.csv"), index=False)

sys.exit(0)



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd

from activitysim.core import config, expressions, inject, pipeline, simulate, tracing
from activitysim.core.util import assign_in_place, reindex

from .util import estimation, trip

logger = logging.getLogger(__name__)


@inject.step()
def stop_frequency(
    tours, tours_merged, stop_frequency_alts, network_los, chunk_size, trace_hh_id
):
    """
    stop frequency model

    For each tour, shoose a number of intermediate inbound stops and outbound stops.
    Create a trip table with inbound and outbound trips.

    Thus, a tour with stop_frequency '2out_0in' will have two outbound and zero inbound stops,
    and four corresponding trips: three outbound, and one inbound.

    Adds stop_frequency str column to trips, with fields

    creates trips table with columns:

    ::

        - person_id
        - household_id
        - tour_id
        - primary_purpose
        - atwork
        - trip_num
        - outbound
        - trip_count

    """

    trace_label = "stop_frequency"
    model_settings_file_name = "stop_frequency.yaml"

    model_settings = config.read_model_settings(model_settings_file_name)

    tours = tours.to_frame()
    tours_merged = tours_merged.to_frame()
    assert not tours_merged.household_id.isnull().any()
    assert not (tours_merged.origin == -1).any()
    assert not (tours_merged.destination == -1).any()

    nest_spec = config.get_logit_model_settings(model_settings)
    constants = config.get_model_constants(model_settings)

    # - run preprocessor to annotate tours_merged
    preprocessor_settings = model_settings.get("preprocessor", None)
    if preprocessor_settings:

        # hack: preprocessor adds origin column in place if it does not exist already
        assert "origin" in tours_merged
        assert "destination" in tours_merged
        od_skim_stack_wrapper = network_los.get_default_skim_dict().wrap(
            "origin", "destination"
        )
        skims = [od_skim_stack_wrapper]

        locals_dict = {"od_skims": od_skim_stack_wrapper, "network_los": network_los}
        locals_dict.update(constants)

        simulate.set_skim_wrapper_targets(tours_merged, skims)

        # this should be pre-slice as some expressions may count tours by type
        annotations = expressions.compute_columns(
            df=tours_merged,
            model_settings=preprocessor_settings,
            locals_dict=locals_dict,
            trace_label=trace_label,
        )

        assign_in_place(tours_merged, annotations)

    tracing.print_summary(
        "stop_frequency segments", tours_merged.primary_purpose, value_counts=True
    )

    spec_segments = model_settings.get("SPEC_SEGMENTS")
    assert (
        spec_segments is not None
    ), f"SPEC_SEGMENTS setting not found in model settings: {model_settings_file_name}"
    segment_col = model_settings.get("SEGMENT_COL")
    assert (
        segment_col is not None
    ), f"SEGMENT_COL setting not found in model settings: {model_settings_file_name}"

    nest_spec = config.get_logit_model_settings(model_settings)

    choices_list = []
    for segment_settings in spec_segments:

        segment_name = segment_settings[segment_col]
        segment_value = segment_settings[segment_col]

        chooser_segment = tours_merged[tours_merged[segment_col] == segment_value]

        if len(chooser_segment) == 0:
            logging.info(f"{trace_label} skipping empty segment {segment_name}")
            continue

        logging.info(
            f"{trace_label} running segment {segment_name} with {chooser_segment.shape[0]} chooser rows"
        )

        estimator = estimation.manager.begin_estimation(
            model_name=segment_name, bundle_name="stop_frequency"
        )

        segment_spec = simulate.read_model_spec(file_name=segment_settings["SPEC"])
        assert segment_spec is not None, (
            "spec for segment_type %s not found" % segment_name
        )

        coefficients_file_name = segment_settings["COEFFICIENTS"]
        coefficients_df = simulate.read_model_coefficients(
            file_name=coefficients_file_name
        )
        segment_spec = simulate.eval_coefficients(
            segment_spec, coefficients_df, estimator
        )

        if estimator:
            estimator.write_spec(segment_settings, bundle_directory=False)
            estimator.write_model_settings(
                model_settings, model_settings_file_name, bundle_directory=True
            )
            estimator.write_coefficients(coefficients_df, segment_settings)
            estimator.write_choosers(chooser_segment)

            estimator.set_chooser_id(chooser_segment.index.name)

        choices = simulate.simple_simulate(
            choosers=chooser_segment,
            spec=segment_spec,
            nest_spec=nest_spec,
            locals_d=constants,
            chunk_size=chunk_size,
            trace_label=tracing.extend_trace_label(trace_label, segment_name),
            trace_choice_name="stops",
            estimator=estimator,
        )

        # convert indexes to alternative names
        choices = pd.Series(segment_spec.columns[choices.values], index=choices.index)

        if estimator:
            estimator.write_choices(choices)
            choices = estimator.get_survey_values(
                choices, "tours", "stop_frequency"
            )  # override choices
            estimator.write_override_choices(choices)
            estimator.end_estimation()

        choices_list.append(choices)

    choices = pd.concat(choices_list)

    tracing.print_summary("stop_frequency", choices, value_counts=True)

    # add stop_frequency choices to tours table
    assign_in_place(tours, choices.to_frame("stop_frequency"))

    # FIXME should have added this when tours created?
    assert "primary_purpose" not in tours
    if "primary_purpose" not in tours.columns:
        # if not already there, then it will have been added by stop_freq_annotate_tours_preprocessor
        assign_in_place(tours, tours_merged[["primary_purpose"]])

    pipeline.replace_table("tours", tours)

    # create trips table
    trips = trip.initialize_from_tours(tours, stop_frequency_alts)
    pipeline.replace_table("trips", trips)
    tracing.register_traceable_table("trips", trips)
    pipeline.get_rn_generator().add_channel("trips", trips)

    if estimator:
        # make sure they created trips with the expected tour_ids
        columns = ["person_id", "household_id", "tour_id", "outbound"]

        survey_trips = estimation.manager.get_survey_table(table_name="trips")
        different = False
        survey_trips_not_in_trips = survey_trips[~survey_trips.index.isin(trips.index)]
        if len(survey_trips_not_in_trips) > 0:
            print(f"survey_trips_not_in_trips\n{survey_trips_not_in_trips}")
            different = True
        trips_not_in_survey_trips = trips[~trips.index.isin(survey_trips.index)]
        if len(survey_trips_not_in_trips) > 0:
            print(f"trips_not_in_survey_trips\n{trips_not_in_survey_trips}")
            different = True
        assert not different

        survey_trips = estimation.manager.get_survey_values(
            trips, table_name="trips", column_names=columns
        )

        trips_differ = (trips[columns] != survey_trips[columns]).any(axis=1)

        if trips_differ.any():
            print("trips_differ\n%s" % trips_differ)
            print("%s of %s tours differ" % (trips_differ.sum(), len(trips_differ)))
            print("differing survey_trips\n%s" % survey_trips[trips_differ])
            print("differing modeled_trips\n%s" % trips[columns][trips_differ])

        assert not trips_differ.any()

    if trace_hh_id:
        tracing.trace_df(
            tours, label="stop_frequency.tours", slicer="person_id", columns=None
        )

        tracing.trace_df(
            trips, label="stop_frequency.trips", slicer="person_id", columns=None
        )

        tracing.trace_df(annotations, label="stop_frequency.annotations", columns=None)

        tracing.trace_df(
            tours_merged,
            label="stop_frequency.tours_merged",
            slicer="person_id",
            columns=None,
        )



from ez_setup import use_setuptools

use_setuptools()  # nopep8

import os
import re

from setuptools import find_packages, setup

with open(os.path.join("activitysim", "__init__.py")) as f:
    info = re.search(r"__.*", f.read(), re.S)
    exec(info[0])

setup(
    name="activitysim",
    version=__version__,
    description=__doc__,
    author="contributing authors",
    author_email="ben.stabler@rsginc.com",
    license="BSD-3",
    url="https://github.com/activitysim/activitysim",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: BSD License",
    ],
    packages=find_packages(exclude=["*.tests"]),
    include_package_data=True,
    entry_points={"console_scripts": ["activitysim=activitysim.cli.main:main"]},
    install_requires=[
        "pyarrow >= 2.0",
        "numpy >= 1.16.1,<=1.21",
        "openmatrix >= 0.3.4.1",
        "pandas >= 1.1.0",
        "pyyaml >= 5.1",
        "tables >= 3.5.1",
        "cytoolz >= 0.8.1",
        "psutil >= 4.1",
        "requests >= 2.7",
        "numba >= 0.51.2",
        "orca >= 1.6",
        "simwrapper > 1.7",
    ],
)



# ActivitySim
# See full license in LICENSE.txt.
import logging
from collections import OrderedDict

from activitysim.abm.models.util import canonical_ids as cid
from activitysim.core import inject

logger = logging.getLogger(__name__)

"""
When the pipeline is restarted and tables are loaded, we need to know which ones
should be registered as random number generator channels.
"""


@inject.injectable()
def rng_channels():

    return cid.RANDOM_CHANNELS


@inject.injectable()
def traceable_tables():

    # names of all traceable tables ordered by dependency on household_id
    # e.g. 'persons' has to be registered AFTER 'households'

    return cid.TRACEABLE_TABLES


@inject.injectable()
def traceable_table_indexes():
    # traceable_table_indexes is OrderedDict {<index_name>: <table_name>}
    # so we can find first registered table to slice by ref_col
    return OrderedDict()


@inject.injectable()
def traceable_table_ids():
    # traceable_table_ids is dict {<table_name>: [<id>, <id>]}
    return dict()


@inject.injectable()
def canonical_table_index_names():
    # traceable_table_ids is dict {<table_name>: [<id>, <id>]}
    return cid.CANONICAL_TABLE_INDEX_NAMES



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd

from activitysim.abm.models.util.canonical_ids import set_trip_index
from activitysim.core import config, inject
from activitysim.core.util import assign_in_place, reindex

logger = logging.getLogger(__name__)


def failed_trip_cohorts(trips, failed):

    # outbound trips in a tour with a failed outbound trip
    bad_outbound_trips = trips.outbound & (
        trips.tour_id.isin(trips.tour_id[failed & trips.outbound])
    )

    # inbound trips in a tour with a failed inbound trip
    bad_inbound_trips = ~trips.outbound & (
        trips.tour_id.isin(trips.tour_id[failed & ~trips.outbound])
    )

    bad_trips = bad_outbound_trips | bad_inbound_trips

    return bad_trips


def flag_failed_trip_leg_mates(trips_df, col_name):
    """
    set boolean flag column of specified name to identify failed trip leg_mates in place
    """

    failed_trip_leg_mates = (
        failed_trip_cohorts(trips_df, trips_df.failed) & ~trips_df.failed
    )
    trips_df.loc[failed_trip_leg_mates, col_name] = True

    # handle outbound and inbound legs independently
    # for ob in [True, False]:
    #     same_leg = (trips_df.outbound == ob)
    #     # tour_ids of all tours with a failed trip in this (outbound or inbound) leg direction
    #     bad_tours = trips_df.tour_id[trips_df.failed & same_leg].unique()
    #     # not-failed leg_mates of all failed trips in this (outbound or inbound) leg direction
    #     failed_trip_leg_mates = same_leg & (trips_df.tour_id.isin(bad_tours)) & ~trips_df.failed
    #     # set the flag column
    #     trips_df.loc[failed_trip_leg_mates, col_name] = True


def cleanup_failed_trips(trips):
    """
    drop failed trips and cleanup fields in leg_mates:

    trip_num        assign new ordinal trip num after failed trips are dropped
    trip_count      assign new count of trips in leg, sans failed trips
    first           update first flag as we may have dropped first trip (last trip can't fail)
    next_trip_id    assign id of next trip in leg after failed trips are dropped
    """

    if trips.failed.any():
        logger.warning(
            "cleanup_failed_trips dropping %s failed trips" % trips.failed.sum()
        )

        trips["patch"] = False
        flag_failed_trip_leg_mates(trips, "patch")

        # drop the original failures
        trips = trips[~trips.failed]

        # increasing trip_id order
        patch_trips = trips[trips.patch].sort_index()

        # recompute fields dependent on trip_num sequence
        grouped = patch_trips.groupby(["tour_id", "outbound"])
        patch_trips["trip_num"] = grouped.cumcount() + 1
        # FIXME - 'clever' hack to avoid regroup - implementation dependent optimization that could change
        patch_trips["trip_count"] = patch_trips["trip_num"] + grouped.cumcount(
            ascending=False
        )

        assign_in_place(trips, patch_trips[["trip_num", "trip_count"]])

        del trips["patch"]

    del trips["failed"]

    return trips


def generate_alternative_sizes(max_duration, max_trips):
    """
    Builds a lookup Numpy array pattern sizes based on the
    number of trips in the leg and the duration available
    to the leg.
    :param max_duration:
    :param max_trips:
    :return:
    """

    def np_shift(xs, n, fill_zero=True):
        if n >= 0:
            shift_array = np.concatenate((np.full(n, np.nan), xs[:-n]))
        else:
            shift_array = np.concatenate((xs[-n:], np.full(-n, np.nan)))
        return (
            np.nan_to_num(shift_array, np.nan).astype(int) if fill_zero else shift_array
        )

    levels = np.empty([max_trips, max_duration + max_trips])
    levels[0] = np.arange(1, max_duration + max_trips + 1)

    for level in np.arange(1, max_trips):
        levels[level] = np_shift(
            np.cumsum(np_shift(levels[level - 1], 1)), -1, fill_zero=False
        )

    return levels[:, : max_duration + 1].astype(int)


def get_time_windows(residual, level):
    """

    :param residual:
    :param level:
    :return:
    """
    ranges = []

    for a in np.arange(residual + 1):
        if level > 1:
            windows = get_time_windows(residual - a, level - 1)
            width_dim = len(windows.shape) - 1
            ranges.append(np.vstack([np.repeat(a, windows.shape[width_dim]), windows]))
        else:
            return np.arange(residual + 1)
    return np.concatenate(ranges, axis=1)


@inject.injectable()
def stop_frequency_alts():
    # alt file for building trips even though simulation is simple_simulate not interaction_simulate
    file_path = config.config_file_path("stop_frequency_alternatives.csv")
    df = pd.read_csv(file_path, comment="#")
    df.set_index("alt", inplace=True)
    return df


def initialize_from_tours(tours, stop_frequency_alts, addtl_tour_cols_to_preserve=None):
    """
    Instantiates a trips table based on tour-level attributes: stop frequency,
    tour origin, tour destination.
    """

    OUTBOUND_ALT = "out"
    assert OUTBOUND_ALT in stop_frequency_alts.columns

    # get the actual alternatives for each person - have to go back to the
    # stop_frequency_alts dataframe to get this - the stop_frequency choice
    # column has the index values for the chosen alternative

    trips = stop_frequency_alts.loc[tours.stop_frequency]

    # assign unique tour identifiers to trips. NOTE: shouldn't be
    # necessary in most cases but there are some edge cases (e.g.
    # tour mode choice logsums) where tour_id will not be unique
    unique_tours = tours.copy().reset_index()
    trips.index = unique_tours.index

    """

    ::

      tours.stop_frequency    =>    proto trips table
      ________________________________________________________
                stop_frequency      |                out  in
      tour_id                       |     tour_id
      954910          1out_1in      |     954910       1   1
      985824          0out_1in      |     985824       0   1
    """

    # reformat with the columns given below
    trips = trips.stack().reset_index()
    trips.columns = ["tour_temp_index", "direction", "trip_count"]

    # tours legs have one more trip than stop
    trips.trip_count += 1

    # prefer direction as boolean
    trips["outbound"] = trips.direction == OUTBOUND_ALT

    """
           tour_temp_index direction  trip_count  outbound
    0             0           out           2         True
    1             0            in           1        False
    2             1           out           2         True
    3             1            in           3        False
    """

    # now do a repeat and a take, so if you have two trips of given type you
    # now have two rows, and zero trips yields zero rows
    trips = trips.take(np.repeat(trips.index.values, trips.trip_count.values))
    trips = trips.reset_index(drop=True)

    grouped = trips.groupby(["tour_temp_index", "outbound"])
    trips["trip_num"] = grouped.cumcount() + 1

    trips["person_id"] = reindex(unique_tours.person_id, trips.tour_temp_index)
    trips["household_id"] = reindex(unique_tours.household_id, trips.tour_temp_index)
    trips["primary_purpose"] = reindex(
        unique_tours.primary_purpose, trips.tour_temp_index
    )

    if addtl_tour_cols_to_preserve is None:
        addtl_tour_cols_to_preserve = []
    for col in addtl_tour_cols_to_preserve:
        trips[col] = reindex(unique_tours[col], trips.tour_temp_index)

    # reorder columns and drop 'direction'
    trips = trips[
        [
            "person_id",
            "household_id",
            "tour_temp_index",
            "primary_purpose",
            "trip_num",
            "outbound",
            "trip_count",
        ]
        + addtl_tour_cols_to_preserve
    ]

    """
      person_id  household_id  tour_temp_index  primary_purpose trip_num  outbound  trip_count
    0     32927         32927        0             work            1        True           2
    1     32927         32927        0             work            2        True           2
    2     32927         32927        0             work            1       False           2
    3     32927         32927        0             work            2       False           2
    4     33993         33993        1             univ            1        True           1
    5     33993         33993        1             univ            1       False           2
    6     33993         33993        1             univ            2       False           2

    """

    # previously in trip_destination.py
    tour_destination = reindex(unique_tours.destination, trips.tour_temp_index).astype(
        np.int64
    )
    tour_origin = reindex(unique_tours.origin, trips.tour_temp_index).astype(np.int64)
    trips["destination"] = np.where(trips.outbound, tour_destination, tour_origin)
    trips["origin"] = np.where(trips.outbound, tour_origin, tour_destination)
    trips["failed"] = False

    # replace temp tour identifier with tour_id
    trips["tour_id"] = reindex(unique_tours.tour_id, trips.tour_temp_index)

    # trip ids are generated based on unique combination of `tour_id`, `outbound`,
    # and `trip_num`. When pseudo-trips are generated from pseudo-tours for the
    # purposes of computing logsums, `tour_id` won't be unique on `outbound` and
    # `trip_num`, so we use `tour_temp_index` instead. this will only be the case
    # when generating temporary pseudo-trips which won't get saved as outputs.
    if (
        trips.groupby(["tour_id", "outbound", "trip_num"])["person_id"].count().max()
        > 1
    ):
        trip_index_tour_id = "tour_temp_index"
    else:
        trip_index_tour_id = "tour_id"

    set_trip_index(trips, trip_index_tour_id)
    del trips["tour_temp_index"]

    return trips



# ActivitySim
# See full license in LICENSE.txt.
import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest

from activitysim.core import random


def test_basic():

    rng = random.Random()

    rng.set_base_seed(0)

    rng.begin_step("test_step")

    global_rng = rng.get_global_rng()

    npt.assert_almost_equal(global_rng.rand(1), [0.8994663])

    # second call should return something different
    with pytest.raises(AssertionError) as excinfo:
        npt.assert_almost_equal(global_rng.rand(1), [0.8994663])
    assert "Arrays are not almost equal" in str(excinfo.value)

    # second call should return something different
    with pytest.raises(RuntimeError) as excinfo:
        rng.set_base_seed(1)
    assert "call set_base_seed before the first step" in str(excinfo.value)


def test_channel():

    channels = {
        "households": "household_id",
        "persons": "person_id",
    }
    rng = random.Random()

    persons = pd.DataFrame(
        {
            "household_id": [1, 1, 2, 2, 2],
        },
        index=[1, 2, 3, 4, 5],
    )
    persons.index.name = "person_id"

    households = pd.DataFrame(
        {
            "data": [1, 1, 2, 2, 2],
        },
        index=[1, 2, 3, 4, 5],
    )
    households.index.name = "household_id"

    rng.begin_step("test_step")

    rng.add_channel("persons", persons)
    rng.add_channel("households", households)

    rands = rng.random_for_df(persons)

    print("rands", np.asanyarray(rands).flatten())

    assert rands.shape == (5, 1)
    test1_expected_rands = [0.1733218, 0.1255693, 0.7384256, 0.3485183, 0.9012387]
    npt.assert_almost_equal(np.asanyarray(rands).flatten(), test1_expected_rands)

    # second call should return something different
    rands = rng.random_for_df(persons)
    test1_expected_rands2 = [0.9105223, 0.5718418, 0.7222742, 0.9062284, 0.3929369]
    npt.assert_almost_equal(np.asanyarray(rands).flatten(), test1_expected_rands2)

    rng.end_step("test_step")

    rng.begin_step("test_step2")

    rands = rng.random_for_df(households)
    expected_rands = [0.417278, 0.2994774, 0.8653719, 0.4429748, 0.5101697]
    npt.assert_almost_equal(np.asanyarray(rands).flatten(), expected_rands)

    choices = rng.choice_for_df(households, [1, 2, 3, 4], 2, replace=True)
    expected_choices = [2, 1, 3, 3, 4, 2, 4, 1, 4, 1]
    npt.assert_almost_equal(choices, expected_choices)

    # should be DIFFERENT the second time
    choices = rng.choice_for_df(households, [1, 2, 3, 4], 2, replace=True)
    expected_choices = [3, 1, 4, 3, 3, 2, 2, 1, 4, 2]
    npt.assert_almost_equal(choices, expected_choices)

    rng.end_step("test_step2")

    rng.begin_step("test_step3")

    rands = rng.random_for_df(households, n=2)

    expected_rands = [
        0.3157928,
        0.3321823,
        0.5194067,
        0.9340083,
        0.9002048,
        0.8754209,
        0.3898816,
        0.4101094,
        0.7351484,
        0.1741092,
    ]

    npt.assert_almost_equal(np.asanyarray(rands).flatten(), expected_rands)

    rng.end_step("test_step3")

    # if we use the same step name a second time, we should get the same results as before
    rng.begin_step("test_step")

    rands = rng.random_for_df(persons)

    print("rands", np.asanyarray(rands).flatten())
    npt.assert_almost_equal(np.asanyarray(rands).flatten(), test1_expected_rands)

    rands = rng.random_for_df(persons)
    npt.assert_almost_equal(np.asanyarray(rands).flatten(), test1_expected_rands2)

    rng.end_step("test_step")



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd

from activitysim.core import (
    chunk,
    config,
    expressions,
    inject,
    logit,
    los,
    mem,
    simulate,
)
from activitysim.core import timetable as tt
from activitysim.core import tracing
from activitysim.core.interaction_sample_simulate import interaction_sample_simulate
from activitysim.core.pathbuilder import TransitVirtualPathBuilder
from activitysim.core.util import reindex

logger = logging.getLogger(__name__)

TDD_CHOICE_COLUMN = "tdd"
USE_BRUTE_FORCE_TO_COMPUTE_LOGSUMS = False

RUN_ALTS_PREPROCESSOR_BEFORE_MERGE = True  # see FIXME below before changing this


def skims_for_logsums(tour_purpose, model_settings, trace_label):

    assert "LOGSUM_SETTINGS" in model_settings

    network_los = inject.get_injectable("network_los")

    skim_dict = network_los.get_default_skim_dict()

    orig_col_name = "home_zone_id"

    destination_for_tour_purpose = model_settings.get("DESTINATION_FOR_TOUR_PURPOSE")
    if isinstance(destination_for_tour_purpose, str):
        dest_col_name = destination_for_tour_purpose
    elif isinstance(destination_for_tour_purpose, dict):
        dest_col_name = destination_for_tour_purpose.get(tour_purpose)
    else:
        raise RuntimeError(
            f"expected string or dict DESTINATION_FOR_TOUR_PURPOSE model_setting for {tour_purpose}"
        )

    odt_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=orig_col_name, dest_key=dest_col_name, dim3_key="out_period"
    )
    dot_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=dest_col_name, dest_key=orig_col_name, dim3_key="in_period"
    )
    odr_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=orig_col_name, dest_key=dest_col_name, dim3_key="in_period"
    )
    dor_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=dest_col_name, dest_key=orig_col_name, dim3_key="out_period"
    )
    od_skim_stack_wrapper = skim_dict.wrap(orig_col_name, dest_col_name)

    skims = {
        "odt_skims": odt_skim_stack_wrapper,
        "dot_skims": dot_skim_stack_wrapper,
        "odr_skims": odr_skim_stack_wrapper,
        "dor_skims": dor_skim_stack_wrapper,
        "od_skims": od_skim_stack_wrapper,
        "orig_col_name": orig_col_name,
        "dest_col_name": dest_col_name,
    }

    if network_los.zone_system == los.THREE_ZONE:
        # fixme - is this a lightweight object?
        tvpb = network_los.tvpb

        tvpb_logsum_odt = tvpb.wrap_logsum(
            orig_key=orig_col_name,
            dest_key=dest_col_name,
            tod_key="out_period",
            segment_key="demographic_segment",
            trace_label=trace_label,
            tag="tvpb_logsum_odt",
        )
        tvpb_logsum_dot = tvpb.wrap_logsum(
            orig_key=dest_col_name,
            dest_key=orig_col_name,
            tod_key="in_period",
            segment_key="demographic_segment",
            trace_label=trace_label,
            tag="tvpb_logsum_dot",
        )

        skims.update(
            {"tvpb_logsum_odt": tvpb_logsum_odt, "tvpb_logsum_dot": tvpb_logsum_dot}
        )

    return skims


def _compute_logsums(
    alt_tdd, tours_merged, tour_purpose, model_settings, network_los, skims, trace_label
):
    """
    compute logsums for tours using skims for alt_tdd out_period and in_period
    """

    trace_label = tracing.extend_trace_label(trace_label, "logsums")

    with chunk.chunk_log(trace_label):
        logsum_settings = config.read_model_settings(model_settings["LOGSUM_SETTINGS"])
        choosers = alt_tdd.join(tours_merged, how="left", rsuffix="_chooser")
        logger.info(
            f"{trace_label} compute_logsums for {choosers.shape[0]} choosers {alt_tdd.shape[0]} alts"
        )

        # - locals_dict
        constants = config.get_model_constants(logsum_settings)
        locals_dict = {}
        locals_dict.update(constants)

        if network_los.zone_system == los.THREE_ZONE:
            # TVPB constants can appear in expressions
            locals_dict.update(
                network_los.setting("TVPB_SETTINGS.tour_mode_choice.CONSTANTS")
            )

        locals_dict.update(skims)

        # constrained coefficients can appear in expressions
        coefficients = simulate.get_segment_coefficients(logsum_settings, tour_purpose)
        locals_dict.update(coefficients)

        # - run preprocessor to annotate choosers
        # allow specification of alternate preprocessor for nontour choosers
        preprocessor = model_settings.get("LOGSUM_PREPROCESSOR", "preprocessor")
        preprocessor_settings = logsum_settings[preprocessor]

        if preprocessor_settings:

            simulate.set_skim_wrapper_targets(choosers, skims)

            expressions.assign_columns(
                df=choosers,
                model_settings=preprocessor_settings,
                locals_dict=locals_dict,
                trace_label=trace_label,
            )

        # - compute logsums
        logsum_spec = simulate.read_model_spec(file_name=logsum_settings["SPEC"])
        logsum_spec = simulate.eval_coefficients(
            logsum_spec, coefficients, estimator=None
        )

        nest_spec = config.get_logit_model_settings(logsum_settings)
        nest_spec = simulate.eval_nest_coefficients(
            nest_spec, coefficients, trace_label
        )

        logsums = simulate.simple_simulate_logsums(
            choosers,
            logsum_spec,
            nest_spec,
            skims=skims,
            locals_d=locals_dict,
            chunk_size=0,
            trace_label=trace_label,
        )

    return logsums


def dedupe_alt_tdd(alt_tdd, tour_purpose, trace_label):

    tdd_segments = inject.get_injectable("tdd_alt_segments", None)
    alt_tdd_periods = None

    logger.info(f"tdd_alt_segments specified for representative logsums")

    with chunk.chunk_log(tracing.extend_trace_label(trace_label, "dedupe_alt_tdd")):

        if tdd_segments is not None:

            dedupe_columns = ["out_period", "in_period"]

            # tdd_alt_segments is optionally segmented by tour purpose
            if "tour_purpose" in tdd_segments:

                is_tdd_for_tour_purpose = tdd_segments.tour_purpose == tour_purpose
                if not is_tdd_for_tour_purpose.any():
                    is_tdd_for_tour_purpose = tdd_segments.tour_purpose.isnull()
                assert (
                    is_tdd_for_tour_purpose.any()
                ), f"no segments found for tour purpose {tour_purpose} in tour_departure_and_duration_segments"

                tdd_segments = tdd_segments[is_tdd_for_tour_purpose].drop(
                    columns=["tour_purpose"]
                )
                assert (
                    len(tdd_segments) > 0
                ), f"tour_purpose '{tour_purpose}' not in tdd_alt_segments"

            # left join representative start on out_period
            alt_tdd_periods = pd.merge(
                alt_tdd[["out_period", "in_period"]].reset_index(),
                tdd_segments[["time_period", "start"]].rename(
                    columns={"time_period": "out_period"}
                ),
                how="left",
                on="out_period",
            )
            chunk.log_df(trace_label, "alt_tdd_periods", alt_tdd_periods)

            # left join representative end on in_period
            alt_tdd_periods = pd.merge(
                alt_tdd_periods,
                tdd_segments[["time_period", "end"]].rename(
                    columns={"time_period": "in_period"}
                ),
                how="left",
                on=["in_period"],
            )
            chunk.log_df(trace_label, "alt_tdd_periods", alt_tdd_periods)

            if tdd_segments.start.isnull().any():
                missing_periods = tdd_segments.out_period[
                    tdd_segments.start.isnull()
                ].unique()
                logger.warning(
                    f"missing out_periods in tdd_alt_segments: {missing_periods}"
                )

            if tdd_segments.end.isnull().any():
                missing_periods = tdd_segments.in_period[
                    tdd_segments.end.isnull()
                ].unique()
                logger.warning(
                    f"missing in_periods in tdd_alt_segments: {missing_periods}"
                )

            assert not tdd_segments.start.isnull().any()
            assert not tdd_segments.end.isnull().any()

            # drop duplicates
            alt_tdd_periods = alt_tdd_periods.drop_duplicates().set_index(
                alt_tdd.index.name
            )
            chunk.log_df(trace_label, "alt_tdd_periods", alt_tdd_periods)

            # representative duration
            alt_tdd_periods["duration"] = (
                alt_tdd_periods["end"] - alt_tdd_periods["start"]
            )
            chunk.log_df(trace_label, "alt_tdd_periods", alt_tdd_periods)

            logger.debug(
                f"{trace_label} "
                f"dedupe_alt_tdd.tdd_alt_segments reduced number of rows by "
                f"{round(100 * (len(alt_tdd) - len(alt_tdd_periods)) / len(alt_tdd), 2)}% "
                f"from {len(alt_tdd)} to {len(alt_tdd_periods)}"
            )

        # if there is no tdd_alt_segments file, we can at least dedupe on 'out_period', 'in_period', 'duration'
        if alt_tdd_periods is None:

            # FIXME This won't work if they reference start or end in logsum calculations
            # for MTC only duration is used (to calculate all_day parking cost)
            dedupe_columns = ["out_period", "in_period", "duration"]

            logger.warning(
                f"No tdd_alt_segments for representative logsums so fallback to "
                f"deduping tdd_alts by time_period and duration"
            )

            # - get list of unique (tour_id, out_period, in_period, duration) in alt_tdd_periods
            # we can cut the number of alts roughly in half (for mtctm1) by conflating duplicates
            alt_tdd_periods = (
                alt_tdd[dedupe_columns]
                .reset_index()
                .drop_duplicates()
                .set_index(alt_tdd.index.name)
            )
            chunk.log_df(trace_label, "alt_tdd_periods", alt_tdd_periods)

            logger.debug(
                f"{trace_label} "
                f"dedupe_alt_tdd.drop_duplicates reduced number of rows by "
                f"{round(100 * (len(alt_tdd) - len(alt_tdd_periods)) / len(alt_tdd), 2)}% "
                f"from {len(alt_tdd)} to {len(alt_tdd_periods)}"
            )

    return alt_tdd_periods, dedupe_columns


def compute_logsums(
    alt_tdd, tours_merged, tour_purpose, model_settings, skims, trace_label
):
    """
    Compute logsums for the tour alt_tdds, which will differ based on their different start, stop
    times of day, which translate to different odt_skim out_period and in_periods.

    In mtctm1, tdds are hourly, but there are only 5 skim time periods, so some of the tdd_alts
    will be the same, once converted to skim time periods. With 5 skim time periods there are
    15 unique out-out period pairs but 190 tdd alternatives.

    For efficiency, rather compute a lot of redundant logsums, we compute logsums for the unique
    (out-period, in-period) pairs and then join them back to the alt_tdds.
    """

    trace_label = tracing.extend_trace_label(trace_label, "compute_logsums")
    network_los = inject.get_injectable("network_los")

    # - in_period and out_period
    assert "out_period" not in alt_tdd
    assert "in_period" not in alt_tdd
    alt_tdd["out_period"] = network_los.skim_time_period_label(alt_tdd["start"])
    alt_tdd["in_period"] = network_los.skim_time_period_label(alt_tdd["end"])
    alt_tdd["duration"] = alt_tdd["end"] - alt_tdd["start"]

    # outside chunk_log context because we extend log_df call for alt_tdd made by our only caller _schedule_tours
    chunk.log_df(trace_label, "alt_tdd", alt_tdd)

    with chunk.chunk_log(trace_label):

        if USE_BRUTE_FORCE_TO_COMPUTE_LOGSUMS:
            # compute logsums for all the tour alt_tdds (inefficient)
            logsums = _compute_logsums(
                alt_tdd,
                tours_merged,
                tour_purpose,
                model_settings,
                network_los,
                skims,
                trace_label,
            )
            return logsums

        index_name = alt_tdd.index.name
        deduped_alt_tdds, redupe_columns = dedupe_alt_tdd(
            alt_tdd, tour_purpose, trace_label
        )
        chunk.log_df(trace_label, "deduped_alt_tdds", deduped_alt_tdds)

        logger.info(
            f"{trace_label} compute_logsums "
            f"deduped_alt_tdds reduced number of rows by "
            f"{round(100 * (len(alt_tdd) - len(deduped_alt_tdds)) / len(alt_tdd), 2)}% "
            f"from {len(alt_tdd)} to {len(deduped_alt_tdds)} compared to USE_BRUTE_FORCE_TO_COMPUTE_LOGSUMS"
        )

        t0 = tracing.print_elapsed_time()

        # - compute logsums for the alt_tdd_periods
        deduped_alt_tdds["logsums"] = _compute_logsums(
            deduped_alt_tdds,
            tours_merged,
            tour_purpose,
            model_settings,
            network_los,
            skims,
            trace_label,
        )

        # tracing.log_runtime(model_name=trace_label, start_time=t0)

        # redupe - join the alt_tdd_period logsums to alt_tdd to get logsums for alt_tdd
        logsums = (
            pd.merge(
                alt_tdd.reset_index(),
                deduped_alt_tdds.reset_index(),
                on=[index_name] + redupe_columns,
                how="left",
            )
            .set_index(index_name)
            .logsums
        )
        chunk.log_df(trace_label, "logsums", logsums)

        del deduped_alt_tdds
        chunk.log_df(trace_label, "deduped_alt_tdds", None)

        # this is really expensive
        TRACE = False
        if TRACE:
            trace_logsums_df = logsums.to_frame("representative_logsum")
            trace_logsums_df["brute_force_logsum"] = _compute_logsums(
                alt_tdd,
                tours_merged,
                tour_purpose,
                model_settings,
                network_los,
                skims,
                trace_label,
            )
            tracing.trace_df(
                trace_logsums_df,
                label=tracing.extend_trace_label(trace_label, "representative_logsums"),
                slicer="NONE",
                transpose=False,
            )

    # leave it to our caller to pick up logsums with call to chunk.log_df
    return logsums


def get_previous_tour_by_tourid(
    current_tour_window_ids, previous_tour_by_window_id, alts
):
    """
    Matches current tours with attributes of previous tours for the same
    person.  See the return value below for more information.

    Parameters
    ----------
    current_tour_window_ids : Series
        A Series of parent ids for the tours we're about make the choice for
        - index should match the tours DataFrame.
    previous_tour_by_window_id : Series
        A Series where the index is the parent (window) id and the value is the index
        of the alternatives of the scheduling.
    alts : DataFrame
        The alternatives of the scheduling.

    Returns
    -------
    prev_alts : DataFrame
        A DataFrame with an index matching the CURRENT tours we're making a
        decision for, but with columns from the PREVIOUS tour of the person
        associated with each of the CURRENT tours.  Columns listed in PREV_TOUR_COLUMNS
        from the alternatives will have "_previous" added as a suffix to keep
        differentiated from the current alternatives that will be part of the
        interaction.
    """

    PREV_TOUR_COLUMNS = ["start", "end"]

    previous_tour_by_tourid = previous_tour_by_window_id.loc[current_tour_window_ids]

    previous_tour_by_tourid = alts.loc[previous_tour_by_tourid, PREV_TOUR_COLUMNS]

    previous_tour_by_tourid.index = current_tour_window_ids.index
    previous_tour_by_tourid.columns = [x + "_previous" for x in PREV_TOUR_COLUMNS]

    return previous_tour_by_tourid


def tdd_interaction_dataset(
    tours, alts, timetable, choice_column, window_id_col, trace_label
):
    """
    interaction_sample_simulate expects
    alts index same as choosers (e.g. tour_id)
    name of choice column in alts

    Parameters
    ----------
    tours : pandas DataFrame
        must have person_id column and index on tour_id
    alts : pandas DataFrame
        alts index must be timetable tdd id
    timetable : TimeTable object
    choice_column : str
        name of column to store alt index in alt_tdd DataFrame
        (since alt_tdd is duplicate index on person_id but unique on person_id,alt_id)

    Returns
    -------
    alt_tdd : pandas DataFrame
        columns: start, end , duration, <choice_column>
        index: tour_id


    """

    trace_label = tracing.extend_trace_label(trace_label, "tdd_interaction_dataset")

    with chunk.chunk_log(trace_label):
        alts_ids = np.tile(alts.index, len(tours.index))
        chunk.log_df(trace_label, "alts_ids", alts_ids)

        tour_ids = np.repeat(tours.index, len(alts.index))
        window_row_ids = np.repeat(tours[window_id_col], len(alts.index))

        alt_tdd = alts.take(alts_ids)

        alt_tdd.index = tour_ids
        alt_tdd[window_id_col] = window_row_ids

        # add tdd alternative id
        # by convention, the choice column is the first column in the interaction dataset
        alt_tdd.insert(loc=0, column=choice_column, value=alts_ids)

        # slice out all non-available tours
        available = timetable.tour_available(
            alt_tdd[window_id_col], alt_tdd[choice_column]
        )
        logger.debug(
            f"tdd_interaction_dataset keeping {available.sum()} of ({len(available)}) available alt_tdds"
        )
        assert available.any()

        chunk.log_df(
            trace_label, "alt_tdd", alt_tdd
        )  # catch this before we slice on available

        alt_tdd = alt_tdd[available]

        chunk.log_df(trace_label, "alt_tdd", alt_tdd)

        # FIXME - don't need this any more after slicing
        del alt_tdd[window_id_col]

    return alt_tdd


def run_alts_preprocessor(model_settings, alts, segment, locals_dict, trace_label):
    """
    run preprocessor on alts, as specified by ALTS_PREPROCESSOR in model_settings

    we are agnostic on whether alts are merged or not

    Parameters
    ----------
    model_settings: dict
        yaml model settings file as dict
    alts: pandas.DataFrame
        tdd_alts or tdd_alts merged wiht choosers (we are agnostic)
    segment: string
        segment selector as understood by caller (e.g. logsum_tour_purpose)
    locals_dict: dict
        we let caller worry about what needs to be in it. though actually depends on modelers needs
    trace_label: string

    Returns
    -------
    alts: pandas.DataFrame
        annotated copy of alts
    """

    preprocessor_settings = model_settings.get("ALTS_PREPROCESSOR", {})

    if segment in preprocessor_settings:
        # segmented by logsum_tour_purpose
        preprocessor_settings = preprocessor_settings.get(segment)
        logger.debug(
            f"running ALTS_PREPROCESSOR with spec for {segment}: {preprocessor_settings.get('SPEC')}"
        )
    elif "SPEC" in preprocessor_settings:
        # unsegmented (either because no segmentation, or fallback if settings has generic preprocessor)
        logger.debug(
            f"running ALTS_PREPROCESSOR with unsegmented spec {preprocessor_settings.get('SPEC')}"
        )
    else:
        logger.debug(
            f"skipping alts preprocesser because no ALTS_PREPROCESSOR segment for {segment}"
        )
        preprocessor_settings = None

    if preprocessor_settings:

        logger.debug(
            f"run_alts_preprocessor calling assign_columns for {segment} preprocessor_settings"
        )
        alts = alts.copy()

        expressions.assign_columns(
            df=alts,
            model_settings=preprocessor_settings,
            locals_dict=locals_dict,
            trace_label=trace_label,
        )

    return alts


def _schedule_tours(
    tours,
    persons_merged,
    alts,
    spec,
    logsum_tour_purpose,
    model_settings,
    skims,
    timetable,
    window_id_col,
    previous_tour,
    tour_owner_id_col,
    estimator,
    tour_trace_label,
):
    """
    previous_tour stores values used to add columns that can be used in the spec
    which have to do with the previous tours per person.  Every column in the
    alternatives table is appended with the suffix "_previous" and made
    available.  So if your alternatives table has columns for start and end,
    then start_previous and end_previous will be set to the start and end of
    the most recent tour for a person.  The first time through,
    start_previous and end_previous are undefined, so make sure to protect
    with a tour_num >= 2 in the variable computation.

    Parameters
    ----------
    tours : DataFrame
        chunk of tours to schedule with unique timetable window_id_col
    persons_merged : DataFrame
        DataFrame of persons to be merged with tours containing attributes referenced
        by expressions in spec
    alts : DataFrame
        DataFrame of alternatives which represent all possible time slots.
        tdd_interaction_dataset function will use timetable to filter them to omit
        unavailable alternatives
    spec : DataFrame
        The spec which will be passed to interaction_simulate.
    model_settings : dict
    timetable : TimeTable
        timetable of timewidows for person (or subtour) with rows for tours[window_id_col]
    window_id_col : str
        column name from tours that identifies timetable owner (or None if tours index)
        - person_id for non/mandatory tours
        - parent_tour_id for subtours,
        - None (tours index) for joint_tours since every tour may have different participants)
    previous_tour: Series
        series with value of tdd_alt choice for last previous tour scheduled for
    tour_owner_id_col : str
        column name from tours that identifies 'owner' of this tour
        (person_id for non/mandatory tours, parent_tour_id for subtours,
        household_id for joint_tours)
    tour_trace_label

    Returns
    -------

    """

    logger.info(
        "%s schedule_tours running %d tour choices" % (tour_trace_label, len(tours))
    )

    # merge persons into tours
    # avoid dual suffix for redundant columns names (e.g. household_id) that appear in both
    tours = pd.merge(
        tours,
        persons_merged,
        left_on="person_id",
        right_index=True,
        suffixes=("", "_y"),
    )
    chunk.log_df(tour_trace_label, "tours", tours)

    # - add explicit window_id_col for timetable owner if it is index
    # if no timetable window_id_col specified, then add index as an explicit column
    # (this is not strictly necessary but its presence makes code simpler in several places)
    if window_id_col is None:
        window_id_col = tours.index.name
        tours[window_id_col] = tours.index

    # timetable can't handle multiple tours per window_id
    assert not tours[window_id_col].duplicated().any()

    # - build interaction dataset filtered to include only available tdd alts
    # dataframe columns start, end , duration, person_id, tdd
    # indexed (not unique) on tour_id
    choice_column = TDD_CHOICE_COLUMN
    alt_tdd = tdd_interaction_dataset(
        tours, alts, timetable, choice_column, window_id_col, tour_trace_label
    )
    # print(f"tours {tours.shape} alts {alts.shape}")

    chunk.log_df(tour_trace_label, "alt_tdd", alt_tdd)

    # - add logsums
    if logsum_tour_purpose:
        logsums = compute_logsums(
            alt_tdd, tours, logsum_tour_purpose, model_settings, skims, tour_trace_label
        )
    else:
        logsums = 0
    alt_tdd["mode_choice_logsum"] = logsums

    del logsums
    chunk.log_df(tour_trace_label, "alt_tdd", alt_tdd)

    # - merge in previous tour columns
    # adds start_previous and end_previous, joins on index
    tours = tours.join(
        get_previous_tour_by_tourid(tours[tour_owner_id_col], previous_tour, alts)
    )
    chunk.log_df(tour_trace_label, "tours", tours)

    # - make choices
    locals_d = {"tt": timetable}
    constants = config.get_model_constants(model_settings)
    if constants is not None:
        locals_d.update(constants)

    if not RUN_ALTS_PREPROCESSOR_BEFORE_MERGE:
        # Note: Clint was running alts_preprocessor here on tdd_interaction_dataset instead of on raw (unmerged) alts
        # and he was using logsum_tour_purpose as selector, although logically it should be the spec_segment
        # It just happened to work for prototype_arc.mandatory_tour_scheduling because, in that model, (unlike semcog)
        # logsum_tour_purpose and spec_segments are aligned (both logsums and spec are segmented on work, school, univ)
        # In any case, I don't see any benefit to doing this here - at least not for any existing implementations
        # but if we do, it will require passing spec_segment to schedule_tours  and _schedule_tours
        # or redundently segmenting alts (yuck!) to conform to more granular tour_segmentation (e.g. univ do school)
        spec_segment = (
            logsum_tour_purpose  # FIXME this is not always right - see note above
        )
        alt_tdd = run_alts_preprocessor(
            model_settings, alt_tdd, spec_segment, locals_d, tour_trace_label
        )
        chunk.log_df(tour_trace_label, "alt_tdd", alt_tdd)

    if estimator:
        # write choosers after annotation
        estimator.write_choosers(tours)
        estimator.set_alt_id(choice_column)
        estimator.write_interaction_sample_alternatives(alt_tdd)

    log_alt_losers = config.setting("log_alt_losers", False)

    choices = interaction_sample_simulate(
        tours,
        alt_tdd,
        spec,
        choice_column=choice_column,
        log_alt_losers=log_alt_losers,
        locals_d=locals_d,
        chunk_size=0,
        trace_label=tour_trace_label,
        estimator=estimator,
    )
    chunk.log_df(tour_trace_label, "choices", choices)

    # - update previous_tour and timetable parameters

    # update previous_tour (series with most recent previous tdd choices) with latest values
    previous_tour.loc[tours[tour_owner_id_col]] = choices.values

    # update timetable with chosen tdd footprints
    timetable.assign(tours[window_id_col], choices)

    return choices


def schedule_tours(
    tours,
    persons_merged,
    alts,
    spec,
    logsum_tour_purpose,
    model_settings,
    timetable,
    timetable_window_id_col,
    previous_tour,
    tour_owner_id_col,
    estimator,
    chunk_size,
    tour_trace_label,
    tour_chunk_tag,
):
    """
    chunking wrapper for _schedule_tours

    While interaction_sample_simulate provides chunking support, the merged tours, persons
    dataframe and the tdd_interaction_dataset are very big, so we want to create them inside
    the chunking loop to minimize memory footprint. So we implement the chunking loop here,
    and pass a chunk_size of 0 to interaction_sample_simulate to disable its chunking support.

    """

    if not tours.index.is_monotonic_increasing:
        logger.info("schedule_tours %s tours not monotonic_increasing - sorting df")
        tours = tours.sort_index()

    logger.info(
        "%s schedule_tours running %d tour choices" % (tour_trace_label, len(tours))
    )

    # no more than one tour per timetable_window per call
    if timetable_window_id_col is None:
        assert not tours.index.duplicated().any()
    else:
        assert not tours[timetable_window_id_col].duplicated().any()

    if "LOGSUM_SETTINGS" in model_settings:
        # we need skims to calculate tvpb skim overhead in 3_ZONE systems for use by calc_rows_per_chunk
        skims = skims_for_logsums(logsum_tour_purpose, model_settings, tour_trace_label)
    else:
        skims = None

    result_list = []
    for i, chooser_chunk, chunk_trace_label in chunk.adaptive_chunked_choosers(
        tours, chunk_size, tour_trace_label, tour_chunk_tag
    ):

        choices = _schedule_tours(
            chooser_chunk,
            persons_merged,
            alts,
            spec,
            logsum_tour_purpose,
            model_settings,
            skims,
            timetable,
            timetable_window_id_col,
            previous_tour,
            tour_owner_id_col,
            estimator,
            tour_trace_label=chunk_trace_label,
        )

        result_list.append(choices)

        chunk.log_df(tour_trace_label, f"result_list", result_list)

    # FIXME: this will require 2X RAM
    # if necessary, could append to hdf5 store on disk:
    # http://pandas.pydata.org/pandas-docs/stable/io.html#id2
    if len(result_list) > 1:
        choices = pd.concat(result_list)

    assert len(choices.index == len(tours.index))

    return choices


def vectorize_tour_scheduling(
    tours,
    persons_merged,
    alts,
    timetable,
    tour_segments,
    tour_segment_col,
    model_settings,
    chunk_size=0,
    trace_label=None,
):
    """
    The purpose of this method is fairly straightforward - it takes tours
    and schedules them into time slots.  Alternatives should be specified so
    as to define those time slots (usually with start and end times).

    schedule_tours adds variables that can be used in the spec which have
    to do with the previous tours per person.  Every column in the
    alternatives table is appended with the suffix "_previous" and made
    available.  So if your alternatives table has columns for start and end,
    then start_previous and end_previous will be set to the start and end of
    the most recent tour for a person.  The first time through,
    start_previous and end_previous are undefined, so make sure to protect
    with a tour_num >= 2 in the variable computation.

    FIXME - fix docstring: tour_segments, tour_segment_col

    Parameters
    ----------
    tours : DataFrame
        DataFrame of tours containing tour attributes, as well as a person_id
        column to define the nth tour for each person.
    persons_merged : DataFrame
        DataFrame of persons containing attributes referenced by expressions in spec
    alts : DataFrame
        DataFrame of alternatives which represent time slots.  Will be passed to
        interaction_simulate in batches for each nth tour.
    spec : DataFrame
        The spec which will be passed to interaction_simulate.
        (or dict of specs keyed on tour_type if tour_types is not None)
    model_settings : dict

    Returns
    -------
    choices : Series
        A Series of choices where the index is the index of the tours
        DataFrame and the values are the index of the alts DataFrame.
    timetable : TimeTable
        persons timetable updated with tours (caller should replace_table for it to persist)
    """

    trace_label = tracing.extend_trace_label(trace_label, "vectorize_tour_scheduling")

    assert len(tours.index) > 0
    assert "tour_num" in tours.columns
    assert "tour_type" in tours.columns

    # tours must be scheduled in increasing trip_num order
    # second trip of type must be in group immediately following first
    # this ought to have been ensured when tours are created (tour_frequency.process_tours)
    choice_list = []

    # keep a series of the the most recent tours for each person
    # initialize with first trip from alts
    previous_tour_by_personid = pd.Series(alts.index[0], index=tours.person_id.unique())

    timetable_window_id_col = "person_id"
    tour_owner_id_col = "person_id"
    compute_logsums = "LOGSUM_SETTINGS" in model_settings

    assert isinstance(tour_segments, dict)

    # no more than one tour per person per call to schedule_tours
    # tours must be scheduled in increasing trip_num order
    # second trip of type must be in group immediately following first
    # segregate scheduling by tour_type if multiple specs passed in dict keyed by tour_type

    for tour_num, nth_tours in tours.groupby("tour_num", sort=True):

        tour_trace_label = tracing.extend_trace_label(trace_label, f"tour_{tour_num}")
        tour_chunk_tag = tracing.extend_trace_label(
            trace_label, f"tour_{1 if tour_num == 1 else 'n'}"
        )

        if tour_segment_col is not None:

            for tour_segment_name, tour_segment_info in tour_segments.items():

                segment_trace_label = tracing.extend_trace_label(
                    tour_trace_label, tour_segment_name
                )
                segment_chunk_tag = tracing.extend_trace_label(
                    tour_chunk_tag, tour_segment_name
                )

                # assume segmentation of spec and coefficients are aligned
                spec_segment_name = tour_segment_info.get("spec_segment_name")
                # assume logsum segmentation is same as tours
                logsum_tour_purpose = tour_segment_name if compute_logsums else None

                nth_tours_in_segment = nth_tours[
                    nth_tours[tour_segment_col] == tour_segment_name
                ]
                if nth_tours_in_segment.empty:
                    logger.info("skipping empty segment %s" % tour_segment_name)
                    continue

                if RUN_ALTS_PREPROCESSOR_BEFORE_MERGE:
                    locals_dict = {}
                    alts = run_alts_preprocessor(
                        model_settings,
                        alts,
                        spec_segment_name,
                        locals_dict,
                        tour_trace_label,
                    )

                choices = schedule_tours(
                    nth_tours_in_segment,
                    persons_merged,
                    alts,
                    spec=tour_segment_info["spec"],
                    logsum_tour_purpose=logsum_tour_purpose,
                    model_settings=model_settings,
                    timetable=timetable,
                    timetable_window_id_col=timetable_window_id_col,
                    previous_tour=previous_tour_by_personid,
                    tour_owner_id_col=tour_owner_id_col,
                    estimator=tour_segment_info.get("estimator"),
                    chunk_size=chunk_size,
                    tour_trace_label=segment_trace_label,
                    tour_chunk_tag=segment_chunk_tag,
                )

                choice_list.append(choices)

        else:

            # MTC non_mandatory_tours are not segmented by tour_purpose and do not require logsums
            # FIXME should support logsums?

            assert (
                not compute_logsums
            ), "logsums for unsegmented spec not implemented because not currently needed"
            assert tour_segments.get("spec_segment_name") is None

            choices = schedule_tours(
                nth_tours,
                persons_merged,
                alts,
                spec=tour_segments["spec"],
                logsum_tour_purpose=None,
                model_settings=model_settings,
                timetable=timetable,
                timetable_window_id_col=timetable_window_id_col,
                previous_tour=previous_tour_by_personid,
                tour_owner_id_col=tour_owner_id_col,
                estimator=tour_segments.get("estimator"),
                chunk_size=chunk_size,
                tour_trace_label=tour_trace_label,
                tour_chunk_tag=tour_chunk_tag,
            )

            choice_list.append(choices)

    choices = pd.concat(choice_list)
    return choices


def vectorize_subtour_scheduling(
    parent_tours,
    subtours,
    persons_merged,
    alts,
    spec,
    model_settings,
    estimator,
    chunk_size=0,
    trace_label=None,
):
    """
    Like vectorize_tour_scheduling but specifically for atwork subtours

    subtours have a few peculiarities necessitating separate treatment:

    Timetable has to be initialized to set all timeperiods outside parent tour footprint as
    unavailable. So atwork subtour timewindows are limited to the footprint of the parent work
    tour. And parent_tour_id' column of tours is used instead of parent_id as timetable row_id.

    Parameters
    ----------
    parent_tours : DataFrame
        parent tours of the subtours (because we need to know the tdd of the parent tour to
        assign_subtour_mask of timetable indexed by parent_tour id
    subtours : DataFrame
        atwork subtours to schedule
    persons_merged : DataFrame
        DataFrame of persons containing attributes referenced by expressions in spec
    alts : DataFrame
        DataFrame of alternatives which represent time slots.  Will be passed to
        interaction_simulate in batches for each nth tour.
    spec : DataFrame
        The spec which will be passed to interaction_simulate.
        (all subtours share same spec regardless of subtour type)
    model_settings : dict
    chunk_size
    trace_label

    Returns
    -------
    choices : Series
        A Series of choices where the index is the index of the subtours
        DataFrame and the values are the index of the alts DataFrame.
    """
    if not trace_label:
        trace_label = "vectorize_non_mandatory_tour_scheduling"

    assert len(subtours.index) > 0
    assert "tour_num" in subtours.columns
    assert "tour_type" in subtours.columns

    timetable_window_id_col = "parent_tour_id"
    tour_owner_id_col = "parent_tour_id"
    logsum_tour_purpose = None  # FIXME logsums not currently supported

    # timetable with a window for each parent tour
    parent_tour_windows = tt.create_timetable_windows(parent_tours, alts)
    timetable = tt.TimeTable(parent_tour_windows, alts)

    # mask the periods outside parent tour footprint
    timetable.assign_subtour_mask(parent_tours.tour_id, parent_tours.tdd)

    # print timetable.windows
    """
    [[7 7 7 0 0 0 0 0 0 0 0 7 7 7 7 7 7 7 7 7 7]
     [7 0 0 0 0 0 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7]
     [7 7 7 7 7 0 0 0 0 0 0 0 0 0 0 7 7 7 7 7 7]
     [7 7 0 0 0 0 0 0 0 7 7 7 7 7 7 7 7 7 7 7 7]]
    """

    choice_list = []

    # keep a series of the the most recent tours for each person
    # initialize with first trip from alts
    previous_tour_by_parent_tour_id = pd.Series(
        alts.index[0], index=subtours["parent_tour_id"].unique()
    )

    # tours must be scheduled in increasing trip_num order
    # second trip of type must be in group immediately following first
    # this ought to have been ensured when tours are created (tour_frequency.process_tours)

    for tour_num, nth_tours in subtours.groupby("tour_num", sort=True):

        tour_trace_label = tracing.extend_trace_label(trace_label, f"tour_{tour_num}")
        tour_chunk_tag = tracing.extend_trace_label(
            trace_label, f"tour_{1 if tour_num == 1 else 'n'}"
        )

        # no more than one tour per timetable window per call to schedule_tours
        assert not nth_tours.parent_tour_id.duplicated().any()

        choices = schedule_tours(
            nth_tours,
            persons_merged,
            alts,
            spec,
            logsum_tour_purpose,
            model_settings,
            timetable,
            timetable_window_id_col,
            previous_tour_by_parent_tour_id,
            tour_owner_id_col,
            estimator,
            chunk_size,
            tour_trace_label,
            tour_chunk_tag,
        )

        choice_list.append(choices)

    choices = pd.concat(choice_list)

    # print "\nfinal timetable.windows\n%s" % timetable.windows
    """
    [[7 7 7 0 0 0 0 2 7 7 4 7 7 7 7 7 7 7 7 7 7]
     [7 0 2 7 4 0 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7]
     [7 7 7 7 7 2 4 0 0 0 0 0 0 0 0 7 7 7 7 7 7]
     [7 7 0 2 7 7 4 0 0 7 7 7 7 7 7 7 7 7 7 7 7]]
    """
    # we dont need to call replace_table() for this nonce timetable
    # because subtours are occuring during persons timetable scheduled time

    return choices


def build_joint_tour_timetables(
    joint_tours, joint_tour_participants, persons_timetable, alts
):

    # timetable with a window for each joint tour
    joint_tour_windows_df = tt.create_timetable_windows(joint_tours, alts)
    joint_tour_timetable = tt.TimeTable(joint_tour_windows_df, alts)

    for participant_num, nth_participants in joint_tour_participants.groupby(
        "participant_num", sort=True
    ):

        # nth_participant windows from persons_timetable
        participant_windows = persons_timetable.slice_windows_by_row_id(
            nth_participants.person_id
        )

        # assign them joint_tour_timetable
        joint_tour_timetable.assign_footprints(
            nth_participants.tour_id, participant_windows
        )

    return joint_tour_timetable


def vectorize_joint_tour_scheduling(
    joint_tours,
    joint_tour_participants,
    persons_merged,
    alts,
    persons_timetable,
    spec,
    model_settings,
    estimator,
    chunk_size=0,
    trace_label=None,
):
    """
    Like vectorize_tour_scheduling but specifically for joint tours

    joint tours have a few peculiarities necessitating separate treatment:

    Timetable has to be initialized to set all timeperiods...

    Parameters
    ----------
    tours : DataFrame
        DataFrame of tours containing tour attributes, as well as a person_id
        column to define the nth tour for each person.
    persons_merged : DataFrame
        DataFrame of persons containing attributes referenced by expressions in spec
    alts : DataFrame
        DataFrame of alternatives which represent time slots.  Will be passed to
        interaction_simulate in batches for each nth tour.
    spec : DataFrame
        The spec which will be passed to interaction_simulate.
        (or dict of specs keyed on tour_type if tour_types is not None)
    model_settings : dict

    Returns
    -------
    choices : Series
        A Series of choices where the index is the index of the tours
        DataFrame and the values are the index of the alts DataFrame.
    persons_timetable : TimeTable
        timetable updated with joint tours (caller should replace_table for it to persist)
    """

    trace_label = tracing.extend_trace_label(
        trace_label, "vectorize_joint_tour_scheduling"
    )

    assert len(joint_tours.index) > 0
    assert "tour_num" in joint_tours.columns
    assert "tour_type" in joint_tours.columns

    timetable_window_id_col = None
    tour_owner_id_col = "household_id"
    logsum_tour_purpose = None  # FIXME logsums not currently supported

    choice_list = []

    # keep a series of the the most recent tours for each person
    # initialize with first trip from alts
    previous_tour_by_householdid = pd.Series(
        alts.index[0], index=joint_tours.household_id.unique()
    )

    # tours must be scheduled in increasing trip_num order
    # second trip of type must be in group immediately following first
    # this ought to have been ensured when tours are created (tour_frequency.process_tours)

    # print "participant windows before scheduling\n%s" % \
    #     persons_timetable.slice_windows_by_row_id(joint_tour_participants.person_id)

    for tour_num, nth_tours in joint_tours.groupby("tour_num", sort=True):

        tour_trace_label = tracing.extend_trace_label(trace_label, f"tour_{tour_num}")
        tour_chunk_tag = tracing.extend_trace_label(
            trace_label, f"tour_{1 if tour_num == 1 else 'n'}"
        )

        # no more than one tour per household per call to schedule_tours
        assert not nth_tours.household_id.duplicated().any()

        nth_participants = joint_tour_participants[
            joint_tour_participants.tour_id.isin(nth_tours.index)
        ]

        timetable = build_joint_tour_timetables(
            nth_tours, nth_participants, persons_timetable, alts
        )

        choices = schedule_tours(
            nth_tours,
            persons_merged,
            alts,
            spec,
            logsum_tour_purpose,
            model_settings,
            timetable,
            timetable_window_id_col,
            previous_tour_by_householdid,
            tour_owner_id_col,
            estimator,
            chunk_size,
            tour_trace_label,
            tour_chunk_tag,
        )

        # - update timetables of all joint tour participants
        persons_timetable.assign(
            nth_participants.person_id, reindex(choices, nth_participants.tour_id)
        )

        choice_list.append(choices)

    choices = pd.concat(choice_list)

    return choices



# ActivitySim
# See full license in LICENSE.txt.
import ctypes
import logging
import multiprocessing
import time
from collections import OrderedDict

import numpy as np
import pandas as pd

from activitysim.abm.tables.size_terms import tour_destination_size_terms
from activitysim.core import config, inject, tracing, util

logger = logging.getLogger(__name__)

"""
ShadowPriceCalculator and associated utility methods

See docstrings for documentation on:

update_shadow_prices    how shadow_price coefficients are calculated
synchronize_choices     interprocess communication to compute aggregate modeled_size
check_fit               convergence criteria for shadow_pric iteration

Import concepts and variables:

model_selector: str
    Identifies a specific location choice model (e.g. 'school', 'workplace')
    The various models work similarly, but use different expression files, model settings, etc.

segment: str
    Identifies a specific demographic segment of a model (e.g. 'elementary' segment of 'school')
    Models can have different size term coefficients (in destinatin_choice_size_terms file) and
    different utility coefficients in models's location and location_sample csv expression files

size_table: pandas.DataFrame


"""


"""
Artisanal reverse semaphores to synchronize concurrent access to shared data buffer

we use the first two rows of the final column in numpy-wrapped shared data as 'reverse semaphores'
(they synchronize concurrent access to shared data resource rather than throttling access)

ShadowPriceCalculator.synchronize_choices coordinates access to the global aggregate zone counts
(local_modeled_size summed across all sub-processes) using these two semaphores
(which are really only tuples of indexes of locations in the shared data array.
"""
TALLY_CHECKIN = (0, -1)
TALLY_CHECKOUT = (1, -1)


def size_table_name(model_selector):
    """
    Returns canonical name of injected destination desired_size table

    Parameters
    ----------
    model_selector : str
        e.g. school or workplace

    Returns
    -------
    table_name : str
    """
    return "%s_destination_size" % model_selector


class ShadowPriceCalculator(object):
    def __init__(
        self, model_settings, num_processes, shared_data=None, shared_data_lock=None
    ):
        """

        Presence of shared_data is used as a flag for multiprocessing
        If we are multiprocessing, shared_data should be a multiprocessing.RawArray buffer
        to aggregate modeled_size across all sub-processes, and shared_data_lock should be
        a multiprocessing.Lock object to coordinate access to that buffer.

        Optionally load saved shadow_prices from data_dir if config setting use_shadow_pricing
        and shadow_setting LOAD_SAVED_SHADOW_PRICES are both True

        Parameters
        ----------
        model_settings : dict
        shared_data : multiprocessing.Array or None (if single process)
        shared_data_lock : numpy array wrapping multiprocessing.RawArray or None (if single process)
        """

        self.num_processes = num_processes
        self.use_shadow_pricing = bool(config.setting("use_shadow_pricing"))
        self.saved_shadow_price_file_path = (
            None  # set by read_saved_shadow_prices if loaded
        )

        self.model_selector = model_settings["MODEL_SELECTOR"]

        full_model_run = config.setting("households_sample_size") == 0
        if self.use_shadow_pricing and not full_model_run:
            logger.warning(
                "deprecated combination of use_shadow_pricing and not full_model_run"
            )

        if (self.num_processes > 1) and not config.setting("fail_fast"):
            # if we are multiprocessing, then fail_fast should be true or we will wait forever for failed processes
            logger.warning(
                "deprecated combination of multiprocessing and not fail_fast"
            )
            raise RuntimeError(
                "Shadow pricing requires fail_fast setting in multiprocessing mode"
            )

        self.segment_ids = model_settings["SEGMENT_IDS"]

        # - modeled_size (set by call to set_choices/synchronize_choices)
        self.modeled_size = None

        if self.use_shadow_pricing:
            self.shadow_settings = config.read_model_settings("shadow_pricing.yaml")

            for k in self.shadow_settings:
                logger.debug(
                    "shadow_settings %s: %s" % (k, self.shadow_settings.get(k))
                )

        # - destination_size_table (desired_size)
        self.desired_size = inject.get_table(
            size_table_name(self.model_selector)
        ).to_frame()
        self.desired_size = self.desired_size.sort_index()

        assert (
            self.desired_size.index.is_monotonic_increasing
        ), f"{size_table_name(self.model_selector)} not is_monotonic_increasing"

        # - shared_data
        if shared_data is not None:
            assert shared_data.shape[0] == self.desired_size.shape[0]
            assert (
                shared_data.shape[1] == self.desired_size.shape[1] + 1
            )  # tally column
            assert shared_data_lock is not None
        self.shared_data = shared_data
        self.shared_data_lock = shared_data_lock

        # - load saved shadow_prices (if available) and set max_iterations accordingly
        if self.use_shadow_pricing:
            self.shadow_prices = None
            self.shadow_price_method = self.shadow_settings["SHADOW_PRICE_METHOD"]
            assert self.shadow_price_method in ["daysim", "ctramp"]

            if self.shadow_settings["LOAD_SAVED_SHADOW_PRICES"]:
                # read_saved_shadow_prices logs error and returns None if file not found
                self.shadow_prices = self.read_saved_shadow_prices(model_settings)

            if self.shadow_prices is None:
                self.max_iterations = self.shadow_settings.get("MAX_ITERATIONS", 5)
            else:
                self.max_iterations = self.shadow_settings.get(
                    "MAX_ITERATIONS_SAVED", 1
                )

            # initial_shadow_price if we did not load
            if self.shadow_prices is None:
                # initial value depends on method
                initial_shadow_price = (
                    1.0 if self.shadow_price_method == "ctramp" else 0.0
                )
                self.shadow_prices = pd.DataFrame(
                    data=initial_shadow_price,
                    columns=self.desired_size.columns,
                    index=self.desired_size.index,
                )
        else:
            self.max_iterations = 1

        self.num_fail = pd.DataFrame(index=self.desired_size.columns)
        self.max_abs_diff = pd.DataFrame(index=self.desired_size.columns)
        self.max_rel_diff = pd.DataFrame(index=self.desired_size.columns)

    def read_saved_shadow_prices(self, model_settings):
        """
        Read saved shadow_prices from csv file in data_dir (so-called warm start)
        returns None if no saved shadow price file name specified or named file not found

        Parameters
        ----------
        model_settings : dict

        Returns
        -------
        shadow_prices : pandas.DataFrame or None
        """

        shadow_prices = None

        # - load saved shadow_prices
        saved_shadow_price_file_name = model_settings.get(
            "SAVED_SHADOW_PRICE_TABLE_NAME"
        )
        if saved_shadow_price_file_name:
            # FIXME - where should we look for this file?
            file_path = config.data_file_path(
                saved_shadow_price_file_name, mandatory=False
            )
            if file_path:
                shadow_prices = pd.read_csv(file_path, index_col=0)
                self.saved_shadow_price_file_path = file_path  # informational
                logger.info("loaded saved_shadow_prices from %s" % file_path)
            else:
                logger.warning("Could not find saved_shadow_prices file %s" % file_path)

        return shadow_prices

    def synchronize_choices(self, local_modeled_size):
        """
        We have to wait until all processes have computed choices and aggregated them by segment
        and zone before we can compute global aggregate zone counts (by segment). Since the global
        zone counts are in shared data, we have to coordinate access to the data structure across
        sub-processes.

        Note that all access to self.shared_data has to be protected by acquiring shared_data_lock

        ShadowPriceCalculator.synchronize_choices coordinates access to the global aggregate
        zone counts (local_modeled_size summed across all sub-processes).

        * All processes wait (in case we are iterating) until any stragglers from the previous
          iteration have exited the building. (TALLY_CHECKOUT goes to zero)

        * Processes then add their local counts into the shared_data and increment TALLY_CHECKIN

        * All processes wait until everybody has checked in (TALLY_CHECKIN == num_processes)

        * Processes make local copy of shared_data and check out (increment TALLY_CHECKOUT)

        * first_in process waits until all processes have checked out, then zeros shared_data
          and clears semaphores

        Parameters
        ----------
        local_modeled_size : pandas DataFrame


        Returns
        -------
        global_modeled_size_df : pandas DataFrame
            local copy of shared global_modeled_size data as dataframe
            with same shape and columns as local_modeled_size
        """

        # shouldn't be called if we are not multiprocessing
        assert self.shared_data is not None
        assert self.num_processes > 1

        def get_tally(t):
            with self.shared_data_lock:
                return self.shared_data[t]

        def wait(tally, target):
            while get_tally(tally) != target:
                time.sleep(1)

        # - nobody checks in until checkout clears
        wait(TALLY_CHECKOUT, 0)

        # - add local_modeled_size data, increment TALLY_CHECKIN
        with self.shared_data_lock:
            first_in = self.shared_data[TALLY_CHECKIN] == 0
            # add local data from df to shared data buffer
            # final column is used for tallys, hence the negative index
            # Ellipsis expands : to fill available dims so [..., 0:-1] is the whole array except for the tallys
            self.shared_data[..., 0:-1] += local_modeled_size.values
            self.shared_data[TALLY_CHECKIN] += 1

        # - wait until everybody else has checked in
        wait(TALLY_CHECKIN, self.num_processes)

        # - copy shared data, increment TALLY_CHECKIN
        with self.shared_data_lock:
            logger.info("copy shared_data")
            # numpy array with sum of local_modeled_size.values from all processes
            global_modeled_size_array = self.shared_data[..., 0:-1].copy()
            self.shared_data[TALLY_CHECKOUT] += 1

        # - first in waits until all other processes have checked out, and cleans tub
        if first_in:
            wait(TALLY_CHECKOUT, self.num_processes)
            with self.shared_data_lock:
                # zero shared_data, clear TALLY_CHECKIN, and TALLY_CHECKOUT semaphores
                self.shared_data[:] = 0
            logger.info("first_in clearing shared_data")

        # convert summed numpy array data to conform to original dataframe
        global_modeled_size_df = pd.DataFrame(
            data=global_modeled_size_array,
            index=local_modeled_size.index,
            columns=local_modeled_size.columns,
        )

        return global_modeled_size_df

    def set_choices(self, choices, segment_ids):
        """
        aggregate individual location choices to modeled_size by zone and segment

        Parameters
        ----------
        choices : pandas.Series
            zone id of location choice indexed by person_id
        segment_ids : pandas.Series
            segment id tag for this individual indexed by person_id

        Returns
        -------
        updates self.modeled_size
        """

        modeled_size = pd.DataFrame(index=self.desired_size.index)
        for seg_name in self.desired_size:

            segment_choices = choices[(segment_ids == self.segment_ids[seg_name])]

            modeled_size[seg_name] = segment_choices.value_counts()

        modeled_size = modeled_size.fillna(0).astype(int)

        if self.num_processes == 1:
            # - not multiprocessing
            self.modeled_size = modeled_size
        else:
            # - if we are multiprocessing, we have to aggregate across sub-processes
            self.modeled_size = self.synchronize_choices(modeled_size)

    def check_fit(self, iteration):
        """
        Check convergence criteria fit of modeled_size to target desired_size
        (For multiprocessing, this is global modeled_size summed across processes,
        so each process will independently calculate the same result.)

        Parameters
        ----------
        iteration: int
            iteration number (informational, for num_failand max_diff history columns)

        Returns
        -------
        converged: boolean

        """

        # fixme

        if not self.use_shadow_pricing:
            return False

        assert self.modeled_size is not None
        assert self.desired_size is not None

        # - convergence criteria for check_fit
        # ignore convergence criteria for zones smaller than size_threshold
        size_threshold = self.shadow_settings["SIZE_THRESHOLD"]
        # zone passes if modeled is within percent_tolerance of  desired_size
        percent_tolerance = self.shadow_settings["PERCENT_TOLERANCE"]
        # max percentage of zones allowed to fail
        fail_threshold = self.shadow_settings["FAIL_THRESHOLD"]

        modeled_size = self.modeled_size
        desired_size = self.desired_size

        abs_diff = (desired_size - modeled_size).abs()

        rel_diff = abs_diff / modeled_size

        # ignore zones where desired_size < threshold
        rel_diff.where(desired_size >= size_threshold, 0, inplace=True)

        # ignore zones where rel_diff < percent_tolerance
        rel_diff.where(rel_diff > (percent_tolerance / 100.0), 0, inplace=True)

        self.num_fail["iter%s" % iteration] = (rel_diff > 0).sum()
        self.max_abs_diff["iter%s" % iteration] = abs_diff.max()
        self.max_rel_diff["iter%s" % iteration] = rel_diff.max()

        total_fails = (rel_diff > 0).values.sum()

        # FIXME - should not count zones where desired_size < threshold? (could calc in init)
        max_fail = (fail_threshold / 100.0) * util.iprod(desired_size.shape)

        converged = total_fails <= max_fail

        # for c in desired_size:
        #     print("check_fit %s segment %s" % (self.model_selector, c))
        #     print("  modeled %s" % (modeled_size[c].sum()))
        #     print("  desired %s" % (desired_size[c].sum()))
        #     print("  max abs diff %s" % (abs_diff[c].max()))
        #     print("  max rel diff %s" % (rel_diff[c].max()))

        logger.info(
            "check_fit %s iteration: %s converged: %s max_fail: %s total_fails: %s"
            % (self.model_selector, iteration, converged, max_fail, total_fails)
        )

        # - convergence stats
        if converged or iteration == self.max_iterations:
            logger.info("\nshadow_pricing max_abs_diff\n%s" % self.max_abs_diff)
            logger.info("\nshadow_pricing max_rel_diff\n%s" % self.max_rel_diff)
            logger.info("\nshadow_pricing num_fail\n%s" % self.num_fail)

        return converged

    def update_shadow_prices(self):
        """
        Adjust shadow_prices based on relative values of modeled_size and desired_size.

        This is the heart of the shadow pricing algorithm.

        The presumption is that shadow_price_adjusted_desired_size (along with other attractors)
        is being used in a utility expression in a location choice model. The goal is to get the
        aggregate location modeled size (choice aggregated by model_selector segment and zone) to
        match desired_size. Since the location choice model may not achieve that goal initially,
        we create a 'shadow price' that tweaks the size_term to encourage the aggregate choices to
        approach the desired target desired_sizes.

        shadow_prices is a table of coefficient (for each zone and segment) that is increases or
        decreases the size term according to whether the modelled population is less or greater
        than the desired_size. If too few total choices are made for a particular zone and
        segment, then its shadow_price is increased, if too many, then it is decreased.

        Since the location choice is being made according to a variety of utilities in the
        expression file, whose relative weights are unknown to this algorithm, the choice of
        how to adjust the shadow_price is not completely straightforward. CTRAMP and Daysim use
        different strategies (see below) and there may not be a single method that works best for
        all expression files. This would be a nice project for the mathematically inclined.

        Returns
        -------
        updates self.shadow_prices
        """

        assert self.use_shadow_pricing

        shadow_price_method = self.shadow_settings["SHADOW_PRICE_METHOD"]

        # can't update_shadow_prices until after first iteration
        # modeled_size should have been set by set_choices at end of previous iteration
        assert self.modeled_size is not None
        assert self.desired_size is not None
        assert self.shadow_prices is not None

        if shadow_price_method == "ctramp":
            # - CTRAMP
            """
            if ( modeledDestinationLocationsByDestZone > 0 )
                shadowPrice *= ( scaledSize / modeledDestinationLocationsByDestZone );
            // else
            //    shadowPrice *= scaledSize;
            """
            damping_factor = self.shadow_settings["DAMPING_FACTOR"]
            assert 0 < damping_factor <= 1

            new_scale_factor = self.desired_size / self.modeled_size
            damped_scale_factor = 1 + (new_scale_factor - 1) * damping_factor
            new_shadow_prices = self.shadow_prices * damped_scale_factor

            # following CTRAMP (revised version - with 0 dest zone case lines commented out)
            # avoid zero-divide for 0 modeled_size, by leaving shadow_prices unchanged
            new_shadow_prices.where(
                self.modeled_size > 0, self.shadow_prices, inplace=True
            )

        elif shadow_price_method == "daysim":
            # - Daysim
            """
            if modeled > desired:  # if modeled is too high, increase shadow price
              target = min(
                modeled,
                desired * (1 + percent_tolerance),
                desired + absolute_tolerance)

            if modeled < desired  # modeled is too low, decrease shadow price
              target = max(
                modeled,
                desired * (1 - percentTolerance),
                desired - absoluteTolerance)

            shadow_price = shadow_price + log(np.maximum(target, 0.01) / np.maximum(modeled, 0.01))
            """
            # FIXME should these be the same as PERCENT_TOLERANCE and FAIL_THRESHOLD above?
            absolute_tolerance = self.shadow_settings["DAYSIM_ABSOLUTE_TOLERANCE"]
            percent_tolerance = self.shadow_settings["DAYSIM_PERCENT_TOLERANCE"] / 100.0
            assert 0 <= percent_tolerance <= 1

            target = np.where(
                self.modeled_size > self.desired_size,
                np.minimum(
                    self.modeled_size,
                    np.minimum(
                        self.desired_size * (1 + percent_tolerance),
                        self.desired_size + absolute_tolerance,
                    ),
                ),
                np.maximum(
                    self.modeled_size,
                    np.maximum(
                        self.desired_size * (1 - percent_tolerance),
                        self.desired_size - absolute_tolerance,
                    ),
                ),
            )

            # adjustment = np.log(np.maximum(target, 0.01) / np.maximum(self.modeled_size, 0.01))
            adjustment = np.log(
                np.maximum(target, 0.01) / np.maximum(self.modeled_size, 1)
            )

            new_shadow_prices = self.shadow_prices + adjustment

        else:
            raise RuntimeError("unknown SHADOW_PRICE_METHOD %s" % shadow_price_method)

        # print("\nself.desired_size\n%s" % self.desired_size.head())
        # print("\nself.modeled_size\n%s" % self.modeled_size.head())
        # print("\nprevious shadow_prices\n%s" % self.shadow_prices.head())
        # print("\nnew_shadow_prices\n%s" % new_shadow_prices.head())

        self.shadow_prices = new_shadow_prices

    def dest_size_terms(self, segment):

        assert segment in self.segment_ids

        size_term_adjustment = 1
        utility_adjustment = 0

        if self.use_shadow_pricing:

            shadow_price_method = self.shadow_settings["SHADOW_PRICE_METHOD"]

            if shadow_price_method == "ctramp":
                size_term_adjustment = self.shadow_prices[segment]
            elif shadow_price_method == "daysim":
                utility_adjustment = self.shadow_prices[segment]
            else:
                raise RuntimeError(
                    "unknown SHADOW_PRICE_METHOD %s" % shadow_price_method
                )

        size_terms = pd.DataFrame(
            {
                "size_term": self.desired_size[segment],
                "shadow_price_size_term_adjustment": size_term_adjustment,
                "shadow_price_utility_adjustment": utility_adjustment,
            },
            index=self.desired_size.index,
        )

        assert size_terms.index.is_monotonic_increasing

        return size_terms

    def write_trace_files(self, iteration):
        """
        Write trace files for this iteration
        Writes desired_size, modeled_size, and shadow_prices tables

        Trace file names are tagged with model_selector and iteration number
        (e.g. self.desired_size => shadow_price_school_desired_size_1)

        Parameters
        ----------
        iteration: int
            current iteration to tag trace file
        """
        logger.info("write_trace_files iteration %s" % iteration)
        if iteration == 1:
            # write desired_size only on first iteration, as it doesn't change
            tracing.write_csv(
                self.desired_size,
                "shadow_price_%s_desired_size" % self.model_selector,
                transpose=False,
            )

        tracing.write_csv(
            self.modeled_size,
            "shadow_price_%s_modeled_size_%s" % (self.model_selector, iteration),
            transpose=False,
        )

        if self.use_shadow_pricing:
            tracing.write_csv(
                self.shadow_prices,
                "shadow_price_%s_shadow_prices_%s" % (self.model_selector, iteration),
                transpose=False,
            )


def block_name(model_selector):
    """
    return canonical block name for model_selector

    Ordinarily and ideally this would just be model_selector, but since mp_tasks saves all
    shared data blocks in a common dict to pass to sub-tasks, we want to be able override
    block naming convention to handle any collisions between model_selector names and skim names.
    Until and unless that happens, we just use model_selector name.

    Parameters
    ----------
    model_selector

    Returns
    -------
    block_name : str
        canonical block name
    """
    return model_selector


def buffers_for_shadow_pricing(shadow_pricing_info):
    """
    Allocate shared_data buffers for multiprocess shadow pricing

    Allocates one buffer per model_selector.
    Buffer datatype and shape specified by shadow_pricing_info

    buffers are multiprocessing.Array (RawArray protected by a multiprocessing.Lock wrapper)
    We don't actually use the wrapped version as it slows access down and doesn't provide
    protection for numpy-wrapped arrays, but it does provide a convenient way to bundle
    RawArray and an associated lock. (ShadowPriceCalculator uses the lock to coordinate access to
    the numpy-wrapped RawArray.)

    Parameters
    ----------
    shadow_pricing_info : dict

    Returns
    -------
        data_buffers : dict {<model_selector> : <shared_data_buffer>}
        dict of multiprocessing.Array keyed by model_selector
    """

    dtype = shadow_pricing_info["dtype"]
    block_shapes = shadow_pricing_info["block_shapes"]

    data_buffers = {}
    for block_key, block_shape in block_shapes.items():

        # buffer_size must be int, not np.int64
        buffer_size = util.iprod(block_shape)

        csz = buffer_size * np.dtype(dtype).itemsize
        logger.info(
            "allocating shared shadow pricing buffer %s %s buffer_size %s bytes %s (%s)"
            % (block_key, buffer_size, block_shape, csz, util.GB(csz))
        )

        if np.issubdtype(dtype, np.int64):
            typecode = ctypes.c_int64
        else:
            raise RuntimeError(
                "buffer_for_shadow_pricing unrecognized dtype %s" % dtype
            )

        shared_data_buffer = multiprocessing.Array(typecode, buffer_size)

        logger.info("buffer_for_shadow_pricing added block %s" % block_key)

        data_buffers[block_key] = shared_data_buffer

    return data_buffers


def shadow_price_data_from_buffers(data_buffers, shadow_pricing_info, model_selector):
    """

    Parameters
    ----------
    data_buffers : dict of {<model_selector> : <multiprocessing.Array>}
        multiprocessing.Array is simply a convenient way to bundle Array and Lock
        we extract the lock and wrap the RawArray in a numpy array for convenience in indexing
        The shared data buffer has shape (<num_zones, <num_segments> + 1)
        extra column is for reverse semaphores with TALLY_CHECKIN and TALLY_CHECKOUT
    shadow_pricing_info : dict
        dict of useful info
           dtype: sp_dtype,
           block_shapes : OrderedDict({<model_selector>: <shape tuple>})
           dict mapping model_selector to block shape (including extra column for semaphores)
           e.g. {'school': (num_zones, num_segments + 1)
    model_selector : str
        location type model_selector (e.g. school or workplace)

    Returns
    -------
    shared_data, shared_data_lock
        shared_data : multiprocessing.Array or None (if single process)
        shared_data_lock : numpy array wrapping multiprocessing.RawArray or None (if single process)
    """

    assert type(data_buffers) == dict

    dtype = shadow_pricing_info["dtype"]
    block_shapes = shadow_pricing_info["block_shapes"]

    if model_selector not in block_shapes:
        raise RuntimeError(
            "Model selector %s not in shadow_pricing_info" % model_selector
        )

    if block_name(model_selector) not in data_buffers:
        raise RuntimeError("Block %s not in data_buffers" % block_name(model_selector))

    shape = block_shapes[model_selector]
    data = data_buffers[block_name(model_selector)]

    return np.frombuffer(data.get_obj(), dtype=dtype).reshape(shape), data.get_lock()


def load_shadow_price_calculator(model_settings):
    """
    Initialize ShadowPriceCalculator for model_selector (e.g. school or workplace)

    If multiprocessing, get the shared_data buffer to coordinate global_desired_size
    calculation across sub-processes

    Parameters
    ----------
    model_settings : dict

    Returns
    -------
    spc : ShadowPriceCalculator
    """

    num_processes = inject.get_injectable("num_processes", 1)

    model_selector = model_settings["MODEL_SELECTOR"]

    # - get shared_data from data_buffers (if multiprocessing)
    data_buffers = inject.get_injectable("data_buffers", None)
    if data_buffers is not None:
        logger.info("Using existing data_buffers for shadow_price")

        # - shadow_pricing_info
        shadow_pricing_info = inject.get_injectable("shadow_pricing_info", None)
        assert shadow_pricing_info is not None

        # - extract data buffer and reshape as numpy array
        data, lock = shadow_price_data_from_buffers(
            data_buffers, shadow_pricing_info, model_selector
        )
    else:
        assert num_processes == 1
        data = None  # ShadowPriceCalculator will allocate its own data
        lock = None

    # - ShadowPriceCalculator
    spc = ShadowPriceCalculator(model_settings, num_processes, data, lock)

    return spc


@inject.step()
def add_size_tables():
    """
    inject tour_destination_size_terms tables for each model_selector (e.g. school, workplace)

    Size tables are pandas dataframes with locations counts for model_selector by zone and segment
    tour_destination_size_terms

    if using shadow pricing, we scale size_table counts to sample population
    (in which case, they have to be created while single-process)

    Scaling is problematic as it breaks household result replicability across sample sizes
    It also changes the magnitude of the size terms so if they are used as utilities in
    expression files, their importance will diminish relative to other utilities as the sample
    size decreases.

    Scaling makes most sense for a full sample in conjunction with shadow pricing, where
    shadow prices can be adjusted iteratively to bring modelled counts into line with desired
    (size table) counts.
    """

    use_shadow_pricing = bool(config.setting("use_shadow_pricing"))

    shadow_settings = config.read_model_settings("shadow_pricing.yaml")
    shadow_pricing_models = shadow_settings.get("shadow_pricing_models")

    if shadow_pricing_models is None:
        logger.warning(
            "shadow_pricing_models list not found in shadow_pricing settings"
        )
        return

    # probably ought not scale if not shadow_pricing (breaks partial sample replicability)
    # but this allows compatability with existing CTRAMP behavior...
    scale_size_table = shadow_settings.get("SCALE_SIZE_TABLE", False)

    # shadow_pricing_models is dict of {<model_selector>: <model_name>}
    # since these are scaled to model size, they have to be created while single-process

    for model_selector, model_name in shadow_pricing_models.items():

        model_settings = config.read_model_settings(model_name)

        assert model_selector == model_settings["MODEL_SELECTOR"]

        assert (
            "SEGMENT_IDS" in model_settings
        ), f"missing SEGMENT_IDS setting in {model_name} model_settings"
        segment_ids = model_settings["SEGMENT_IDS"]
        chooser_table_name = model_settings["CHOOSER_TABLE_NAME"]
        chooser_segment_column = model_settings["CHOOSER_SEGMENT_COLUMN_NAME"]

        choosers_df = inject.get_table(chooser_table_name).to_frame()
        if "CHOOSER_FILTER_COLUMN_NAME" in model_settings:
            choosers_df = choosers_df[
                choosers_df[model_settings["CHOOSER_FILTER_COLUMN_NAME"]] != 0
            ]

        # - raw_desired_size
        land_use = inject.get_table("land_use")
        size_terms = inject.get_injectable("size_terms")
        raw_size = tour_destination_size_terms(land_use, size_terms, model_selector)
        assert set(raw_size.columns) == set(segment_ids.keys())

        if use_shadow_pricing or scale_size_table:

            # - scale size_table counts to sample population
            # scaled_size = zone_size * (total_segment_modeled / total_segment_desired)

            # segment scale factor (modeled / desired) keyed by segment_name
            segment_scale_factors = {}
            for c in raw_size:
                # number of zone demographics desired destination choices
                segment_desired_size = raw_size[c].astype(np.float64).sum()

                # number of synthetic population choosers in segment
                segment_chooser_count = (
                    choosers_df[chooser_segment_column] == segment_ids[c]
                ).sum()

                segment_scale_factors[c] = segment_chooser_count / np.maximum(
                    segment_desired_size, 1
                )

                logger.info(
                    "add_desired_size_tables %s segment %s "
                    "desired %s modeled %s scale_factor %s"
                    % (
                        chooser_table_name,
                        c,
                        segment_desired_size,
                        segment_chooser_count,
                        segment_scale_factors[c],
                    )
                )

            # FIXME - should we be rounding?
            scaled_size = (raw_size * segment_scale_factors).round()
        else:
            scaled_size = raw_size

        logger.debug(
            f"add_size_table {size_table_name(model_selector)} ({scaled_size.shape}) for {model_selector}"
        )

        assert (
            scaled_size.index.is_monotonic_increasing
        ), f"size table {size_table_name(model_selector)} not is_monotonic_increasing"

        inject.add_table(size_table_name(model_selector), scaled_size)


def get_shadow_pricing_info():
    """
    return dict with info about dtype and shapes of desired and modeled size tables

    block shape is (num_zones, num_segments + 1)


    Returns
    -------
    shadow_pricing_info: dict
        dtype: <sp_dtype>,
        block_shapes: dict {<model_selector>: <block_shape>}
    """

    land_use = inject.get_table("land_use")
    size_terms = inject.get_injectable("size_terms")

    shadow_settings = config.read_model_settings("shadow_pricing.yaml")

    # shadow_pricing_models is dict of {<model_selector>: <model_name>}
    shadow_pricing_models = shadow_settings.get("shadow_pricing_models", {})

    blocks = OrderedDict()
    for model_selector in shadow_pricing_models:

        sp_rows = len(land_use)
        sp_cols = len(size_terms[size_terms.model_selector == model_selector])

        # extra tally column for TALLY_CHECKIN and TALLY_CHECKOUT semaphores
        blocks[block_name(model_selector)] = (sp_rows, sp_cols + 1)

    sp_dtype = np.int64

    shadow_pricing_info = {
        "dtype": sp_dtype,
        "block_shapes": blocks,
    }

    for k in shadow_pricing_info:
        logger.debug("shadow_pricing_info %s: %s" % (k, shadow_pricing_info.get(k)))

    return shadow_pricing_info


@inject.injectable(cache=True)
def shadow_pricing_info():

    # when multiprocessing with shared data mp_tasks has to call network_los methods
    # get_shadow_pricing_info() and buffers_for_shadow_pricing()
    logger.debug("loading shadow_pricing_info injectable")

    return get_shadow_pricing_info()



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd

from activitysim.core import assign, chunk, config, inject, los, mem, pipeline, tracing
from activitysim.core.pathbuilder import TransitVirtualPathBuilder

logger = logging.getLogger(__name__)


def compute_accessibilities_for_zones(
    accessibility_df,
    land_use_df,
    assignment_spec,
    constants,
    network_los,
    trace_od,
    trace_label,
):

    orig_zones = accessibility_df.index.values
    dest_zones = land_use_df.index.values

    orig_zone_count = len(orig_zones)
    dest_zone_count = len(dest_zones)

    logger.info(
        "Running %s with %d orig zones %d dest zones"
        % (trace_label, orig_zone_count, dest_zone_count)
    )

    # create OD dataframe
    od_df = pd.DataFrame(
        data={
            "orig": np.repeat(orig_zones, dest_zone_count),
            "dest": np.tile(dest_zones, orig_zone_count),
        }
    )

    if trace_od:
        trace_orig, trace_dest = trace_od
        trace_od_rows = (od_df.orig == trace_orig) & (od_df.dest == trace_dest)
    else:
        trace_od_rows = None

    # merge land_use_columns into od_df
    logger.info(f"{trace_label}: merge land_use_columns into od_df")
    od_df = pd.merge(od_df, land_use_df, left_on="dest", right_index=True).sort_index()
    chunk.log_df(trace_label, "od_df", od_df)

    locals_d = {
        "log": np.log,
        "exp": np.exp,
        "network_los": network_los,
    }
    locals_d.update(constants)

    skim_dict = network_los.get_default_skim_dict()
    locals_d["skim_od"] = skim_dict.wrap("orig", "dest").set_df(od_df)
    locals_d["skim_do"] = skim_dict.wrap("dest", "orig").set_df(od_df)

    if network_los.zone_system == los.THREE_ZONE:
        locals_d["tvpb"] = network_los.tvpb

    logger.info(f"{trace_label}: assign.assign_variables")
    results, trace_results, trace_assigned_locals = assign.assign_variables(
        assignment_spec,
        od_df,
        locals_d,
        trace_rows=trace_od_rows,
        trace_label=trace_label,
        chunk_log=True,
    )

    chunk.log_df(trace_label, "results", results)
    logger.info(f"{trace_label}: have results")

    # accessibility_df = accessibility_df.copy()
    for column in results.columns:
        data = np.asanyarray(results[column])
        data.shape = (orig_zone_count, dest_zone_count)  # (o,d)
        accessibility_df[column] = np.log(np.sum(data, axis=1) + 1)

    if trace_od:

        if not trace_od_rows.any():
            logger.warning(
                f"trace_od not found origin = {trace_orig}, dest = {trace_dest}"
            )
        else:

            # add OD columns to trace results
            df = pd.concat([od_df[trace_od_rows], trace_results], axis=1)

            # dump the trace results table (with _temp variables) to aid debugging
            tracing.trace_df(
                df,
                label="accessibility",
                index_label="skim_offset",
                slicer="NONE",
                warn_if_empty=True,
            )

            if trace_assigned_locals:
                tracing.write_csv(
                    trace_assigned_locals, file_name="accessibility_locals"
                )

    return accessibility_df


@inject.step()
def compute_accessibility(land_use, accessibility, network_los, chunk_size, trace_od):

    """
    Compute accessibility for each zone in land use file using expressions from accessibility_spec

    The actual results depend on the expressions in accessibility_spec, but this is initially
    intended to permit implementation of the mtc accessibility calculation as implemented by
    Accessibility.job

    Compute measures of accessibility used by the automobile ownership model.
    The accessibility measure first multiplies an employment variable by a mode-specific decay
    function.  The product reflects the difficulty of accessing the activities the farther
    (in terms of round-trip travel time) the jobs are from the location in question. The products
    to each destination zone are next summed over each origin zone, and the logarithm of the
    product mutes large differences.  The decay function on the walk accessibility measure is
    steeper than automobile or transit.  The minimum accessibility is zero.
    """

    trace_label = "compute_accessibility"
    model_settings = config.read_model_settings("accessibility.yaml")
    assignment_spec = assign.read_assignment_spec(
        config.config_file_path("accessibility.csv")
    )

    accessibility_df = accessibility.to_frame()
    if len(accessibility_df.columns) > 0:
        logger.warning(
            f"accessibility table is not empty. Columns:{list(accessibility_df.columns)}"
        )
        raise RuntimeError(f"accessibility table is not empty.")

    constants = config.get_model_constants(model_settings)

    # only include the land_use columns needed by spec, as specified by land_use_columns model_setting
    land_use_columns = model_settings.get("land_use_columns", [])
    land_use_df = land_use.to_frame()
    land_use_df = land_use_df[land_use_columns]

    logger.info(
        f"Running {trace_label} with {len(accessibility_df.index)} orig zones {len(land_use_df)} dest zones"
    )

    accessibilities_list = []

    for i, chooser_chunk, chunk_trace_label in chunk.adaptive_chunked_choosers(
        accessibility_df, chunk_size, trace_label
    ):

        accessibilities = compute_accessibilities_for_zones(
            chooser_chunk,
            land_use_df,
            assignment_spec,
            constants,
            network_los,
            trace_od,
            trace_label,
        )
        accessibilities_list.append(accessibilities)

    accessibility_df = pd.concat(accessibilities_list)

    logger.info(f"{trace_label} computed accessibilities {accessibility_df.shape}")

    # - write table to pipeline
    pipeline.replace_table("accessibility", accessibility_df)



#  python ~/work/activitysim/activitysim/examples/example_estimation/build_example_data/build_stop_coeffs.py

import numpy as np
import pandas as pd

FIRST_RUN = True

# work, school, univ, social, shopping, eatout, escort,atwork,othmaint,othdiscr
for what in [
    "work",
    "school",
    "univ",
    "social",
    "shopping",
    "eatout",
    "escort",
    "atwork",
    "othmaint",
    "othdiscr",
]:

    if FIRST_RUN:
        df = pd.read_csv(f"stop_frequency_{what}.csv", comment="#")
        df.to_csv(f"stop_frequency_backup_{what}.csv", index=False)
    else:
        df = pd.read_csv(f"stop_frequency_backup_{what}.csv", comment="#")

    del df["Expression"]

    df = df.set_index("Description").unstack()

    # drop empty coefficients
    df = df[~df.isnull()]

    # want index as columns
    df = df.reset_index().rename(columns={"level_0": "alt", 0: "value"})

    # drop duplicate coefficients on same spec row
    df = df[~df[["Description", "value"]].duplicated(keep="first")]

    dupes = df[["Description"]].duplicated(keep=False)
    df["coefficient_name"] = np.where(
        dupes, "coef_" + df.Description + "_" + df["alt"], "coef_" + df.Description
    )
    df["coefficient_name"] = df["coefficient_name"].str.lower()
    df["coefficient_name"] = df["coefficient_name"].str.replace(
        "[^a-zZ-Z0-9]+", "_", regex=True
    )
    del df["alt"]

    df.to_csv(f"stop_frequency_coefficients_{what}.csv", index=False)

    spec = pd.read_csv(f"stop_frequency_backup_{what}.csv", comment="#")

    alt_cols = spec.columns[2:].values

    for index, row in df.iterrows():
        m = {row["value"]: row["coefficient_name"]}
        alts = spec.loc[spec.Description == row["Description"], alt_cols].values[0]
        alts = [m.get(a, a) for a in alts]

        spec.loc[spec.Description == row["Description"], alt_cols] = [
            m.get(a, a) for a in alts
        ]

    spec.insert(loc=0, column="Label", value="util_" + spec.Description)

    spec["Label"] = spec["Label"].str.lower()
    spec["Label"] = spec["Label"].str.replace("[^a-zZ-Z0-9]+", "_", regex=True)

    df.to_csv(f"stop_frequency_coefficients_{what}.csv", index=False)
    spec.to_csv(f"stop_frequency_{what}.csv", index=False)
